/**
  *  \file client/chartloc.h
  *  \brief GChartLocation class
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_CHARTLOC_H
#define PCC_V2_CLIENT_CHARTLOC_H

#include "client/chart.h"
#include "game/objlist.h"
#include "int/keymap.h"
#include "game/objsel.h"

/** Starchart location. Manages a location and viewport for the
    starchart, and a redraw flag.

    The center is always a normalized coordinate, and represents the center
    of the map image.

    The location is an absolute coordinate relative to the center. It may not
    be a normalized coordinate if there is a map border between the center and
    the location. The object list is still built correctly.

    When the center crosses a map border, an appropriate adjustment will be
    computed. The location will automatically be adjusted; users can use the wrap
    adjustment to adjust further coordinates. The wrap adjustment will be reset
    upon markChanged().

    This class also manages zoom operations as well as a scrollable object list.
    All changes are folded into a change flag which can be queried and reset. */
class GChartLocation {
    GChartViewport viewport;    /**< Center of map image, normalized. */
    GPoint         location;    /**< Cursor location. Not necessary normalized. */
    GPoint         wrap_adjust; /**< Wrap-adjustment. Sum of all moves across borders. */
    GObjectList   list;         /**< List of objects at this location. */
    GSimpleObjectSelection list_current;
    int            list_top;    /**< Index (1-based) of top-most object in list. */
    int            pid;         /**< Planet Id at this location. */
    string_t       view_name;   /**< Name of current view. */
    IntKeymapRef   keymap;      /**< Active keymap. */

    bool           changed;     /**< Change flag. */

    void      updateLocation();
    void      updateObject();
    void      clearWrapAdjust();
    void      adjustObjectListTop();
 public:
    GChartLocation();
    ~GChartLocation();

    void         init();
    void         initAfterLoad();
    void         initEntry(const GfxRect& area);

    void         setCenter(GPoint pt);
    GPoint       getCenter() const;
    void         setLocation(GPoint pt);
    GPoint       getLocation() const;
    GPoint       getWrapAdjust() const;

    void         setViewName(string_t name);
    string_t     getViewName() const;
    void         setKeymap(IntKeymapRef keymap);
    IntKeymapRef getKeymap() const;

    int          getCurrentShipId() const;
    GObjectReference getCurrentObject() const;
    void         setCurrentObject(const GObject& obj);

    void         doZoom(bool enlarge);
    void         doScroll(int direction, bool mark);

    GChartViewport& getViewport();

    const GObjectList& getObjectList() const;
    const GObjectSelection& getObjectSelection() const;
    int       getObjectListHeight() const;
    int       getObjectListTop() const;

    void      toggleOptions(int32 which);

    bool      isChanged() const;
    void      markChanged();
    void      clearChanged();
};


inline GPoint
GChartLocation::getCenter() const
{
    return viewport.center;
}

inline GPoint
GChartLocation::getLocation() const
{
    return location;
}

inline void
GChartLocation::setKeymap(IntKeymapRef keymap)
{
    this->keymap = keymap;
}

inline IntKeymapRef
GChartLocation::getKeymap() const
{
    return keymap;
}

inline GChartViewport&
GChartLocation::getViewport()
{
    // cannot be const, because the interesting methods of GChartViewport are not const
    return viewport;
}

inline const GObjectList&
GChartLocation::getObjectList() const
{
    return list;
}

inline const GObjectSelection&
GChartLocation::getObjectSelection() const
{
    return list_current;
}

#endif
