/**
  *  \file client/scr-sim.cc
  *  \brief Battle Simulator Screen
  *
  *  This contains all the widgets and routines for editing a
  *  simulation.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "client/scr-sim.h"

#include "client/chart.h"
#include "client/dialogs/fleetcost.h"
#include "client/dialogs/helpdlg.h"
#include "client/dialogs/hullsel.h"
#include "client/dialogs/simalliances.h"
#include "client/dialogs/simedit.h"
#include "client/dialogs/simflakedit.h"
#include "client/dialogs/simfunction.h"
#include "client/dialogs/simoptions.h"
#include "client/dialogs/simplanetedit.h"
#include "client/dlg-fcode.h"
#include "client/scr-result.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/simlist.h"
#include "client/wtile.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "defs.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/racename.h"
#include "game/spec.h"
#include "game/team.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "io/glob.h"
#include "resmgr/resid.h"
#include "resmgr/resmgr.h"
#include "ui/combobox.h"
#include "ui/commands.h"
#include "ui/dialogs/filedlg.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/table.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/focusgroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/quit.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"

/** List of battle participants, with editing functions. */
class WSimListWithHandler : public WSimList {
 public:
    WSimListWithHandler(GSimulatorGameInterface& game_if, GSimState& state);

    bool handleEvent(const UIEvent& e, bool second_pass);

    void addShip();
    void replicateShip();
    void deleteShip();

    void writeToGame();
    void updateFromGame();
    void switchToGame();

    void editAggressiveness();
    void editType(bool after_add);
    void editFCode();
    void incrFCode();
    void editPrimaryWeapon();
    void editSecondaryWeapon();
    void editName();
    void editId();
    void editDamage();
    void editShield();
    void editMass();
    void editCrew();
    void editExperienceLevel();
    void editOwner();
    void editEngine();
    void editInterceptId();
    void editAbilities();

    void swapUp();
    void swapDown();

    void toggleCloak();
    void toggleRandomFC();
    void toggleEnabled();

    void replicateCurrentShip(int count);
};

class WSimPlanetButton : public UIAbstractButton {
    GSimState& state;
    GSimulatorGameInterface& game_if;
    SignalConnection conn_edit;
    SignalConnection conn_structure;
 public:
    WSimPlanetButton(GSimulatorGameInterface& game_if, GSimState& state);
    ~WSimPlanetButton();
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
};

static UIBaseWidget&
wrapIncrement(ObjectHolder& h, int id, UIWidget& w)
{
    UIButton& btnMore  = h.add(new UIButton("+", '+'));
    UIButton& btnFewer = h.add(new UIButton("-", '-'));
    UIFocusGroup& g = h.add(new UIFocusGroup(UIHBoxLayout::instance5, id, 5));
    g.add(btnFewer);
    g.add(w);
    g.add(btnMore);
    g.setFocusedWidget(w);
    return g;    
}

static string_t
makeDefaultName(int shipId)
{
    return format(_("Ship %d"), shipId);
}

static void
setPlanetName(GSimulatorGameInterface& game_if, GSimPlanet& pl)
{
    pl.setName(game_if.getPlanetName(pl.getId()));
}

static bool
sortById(const GSimShip& a, const GSimShip& b)
{
    return a.getId() < b.getId();
}

static bool
sortByOwner(const GSimShip& a, const GSimShip& b)
{
    if (a.getOwner() == b.getOwner())
        return a.getId() < b.getId();
    else
        return a.getOwner() < b.getOwner();
}

static bool
sortByHull(const GSimShip& a, const GSimShip& b)
{
    if (a.getHull() == b.getHull())
        return a.getId() < b.getId();
    else
        return a.getHull() < b.getHull();
}

static bool
sortByBattleOrder(const GSimShip& a, const GSimShip& b)
{
    if (WSimScreen::sim_opts.vcr_mode == GSimOptions::VcrHost || WSimScreen::sim_opts.vcr_mode == GSimOptions::VcrNuHost) {
        return sortByBattleOrderTHost(&a, &b);
    } else {
        /* sortByBattleOrderPHost takes non-const because it wants to cast them
           around to support planets. It still calls only non-modifying operations. */
        return sortByBattleOrderPHost(const_cast<GSimShip*>(&a), const_cast<GSimShip*>(&b));
    }
}

static bool
sortByName(const GSimShip& a, const GSimShip& b)
{
    const string_t aName = strUCase(a.getName());
    const string_t bName = strUCase(b.getName());
    if (aName == bName)
        return a.getId() < b.getId();
    else
        return aName < bName;
}

static string_t
describeTorpedo(const GTorpLauncher& torp)
{
    int scale = host.isPHost() && config.AllowAlternativeCombat() ? 1 : 2;
    return format(_("%X - %d\t(Tech %d, K%d, D%d)"))
        << torp.getId()
        << torp.getName()
        << torp.getTech()
        << scale*torp.getKill()
        << scale*torp.getDamage();
}

static string_t
describeBeam(const GBeam& beam)
{
    return format(_("%X - %d\t(Tech %d, K%d, D%d)"))
        << beam.getId()
        << beam.getName()
        << beam.getTech()
        << beam.getKill()
        << beam.getDamage();
}

static void
avoidSelfAgression(GSimShip& sh)
{
    if (sh.getAggressiveness() == sh.getOwner()) {
        sh.setAggressiveness(GSimShip::agg_Kill);
        sh.setFlags(sh.getFlags() & ~GSimShip::fl_Cloaked);
    }
}

bool
WSimScreen::isExpGame()
{
    /* This used to check also the host mode. However, it was never implemented
       consistently. Therefore, enable editing experience whenever it is configured,
       even if it's not used in the current simulation. */
    return (config.NumExperienceLevels() > 0);
}

void
WSimScreen::setDefaultRelations(GSimulatorGameInterface& game_if)
{
    game_if.setDefaultRelations(sim_opts.alliance_settings, sim_opts.enemy_settings);
}

/************************** WSimListWithHandler **************************/

/** Constructor.
    \param state [in] Simulation state to edit. */
WSimListWithHandler::WSimListWithHandler(GSimulatorGameInterface& game_if, GSimState& state)
    : WSimList(game_if, state, false)
{ }

bool
WSimListWithHandler::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_INSERT:
            addShip();
            getSimState().doUpdates();
            return true;
         case SDLK_INSERT + ss_Ctrl:
         case '*':
            replicateShip();
            getSimState().doUpdates();
            return true;
         case SDLK_DELETE:
            deleteShip();
            getSimState().doUpdates();
            return true;
         case SDLK_F1:
            switchToGame();
            return true;
         case SDLK_UP + ss_Ctrl:
            swapUp();
            getSimState().doUpdates();
            return true;
         case SDLK_DOWN + ss_Ctrl:
            swapDown();
            getSimState().doUpdates();
            return true;
         case 'u':
            updateFromGame();
            getSimState().doUpdates();
            return true;
         case 'w':
            writeToGame();
            return true;
         case 'a':
            editAggressiveness();
            getSimState().doUpdates();
            return true;
         case 't':
            editType(false);
            getSimState().doUpdates();
            return true;
         case 'f':
            editFCode();
            getSimState().doUpdates();
            return true;
         case 'F':
            incrFCode();
            getSimState().doUpdates();
            return true;
         case '1':
            editPrimaryWeapon();
            getSimState().doUpdates();
            return true;
         case '2':
            editSecondaryWeapon();
            getSimState().doUpdates();
            return true;
         case 'n':
            editName();
            getSimState().doUpdates();
            return true;
         case 'i':
            editId();
            getSimState().doUpdates();
            return true;
         case 'd':
            editDamage();
            getSimState().doUpdates();
            return true;
         case 's':
            editShield();
            getSimState().doUpdates();
            return true;
         case 'm':
            editMass();
            getSimState().doUpdates();
            return true;
         case 'c':
            editCrew();
            getSimState().doUpdates();
            return true;
         case 'l':
            editExperienceLevel();
            getSimState().doUpdates();
            return true;
         case 'o':
            editOwner();
            getSimState().doUpdates();
            return true;
         case 'e':
            editEngine();
            getSimState().doUpdates();
            return true;
         case 'v':
            toggleCloak();
            getSimState().doUpdates();
            return true;
         case 'r':
            toggleRandomFC();
            getSimState().doUpdates();
            return true;
         case '.':
            toggleEnabled();
            getSimState().doUpdates();
            return true;
         case 'x':
            editInterceptId();
            getSimState().doUpdates();
            return true;
         case 'b':
            editAbilities();
            getSimState().doUpdates();
            return true;
         case 'k':
            if (GSimShip* sh = getCurrentShip()) {
                editSimulatorFlakRatings(*sh, WSimScreen::sim_opts);
                getSimState().doUpdates();
            }
            return true;
        }
    }
    return WSimList::handleEvent(e, second_pass);
}

/** Add new ship. */
void
WSimListWithHandler::addShip()
{
    GSimState& st = getSimState();
    if (st.getNumShips() >= st.MAX_SHIPS) {
        messageBox(_("This simulation already contains the maximum possible number of ships."),
                   _("Battle Simulator"));
    } else {
        // Create the ship
        if (st.getNumShips() == 0) {
            st.addShip();
            st.getShip(0).setHull(1);
            st.getShip(0).setId(1);
            st.getShip(0).setAggressiveness(GSimShip::agg_Kill);
            st.getShip(0).setName(makeDefaultName(1));
            st.doUpdates();
            scrollTo(0, 1);
        } else {
            replicateCurrentShip(1);
            st.doUpdates();
        }

        // Mark it enabled, even when created by pressing INS on a disabled ship
        if (GSimShip* sh = getCurrentShip()) {
            sh->setFlags(sh->getFlags() & ~GSimShip::fl_Deactivated);
        }

        // Edit type
        editType(true);
    }
}

/** Replicate ship. User is asked for how often to replicate it. */
void
WSimListWithHandler::replicateShip()
{
    GSimState& st = getSimState();
    if (st.getNumShips() == 0) {
        // nothing there to replicate
    } else if (st.getNumShips() >= st.MAX_SHIPS) {
        messageBox(_("This simulation already contains the maximum possible number of ships."),
                   _("Battle Simulator"));
    } else {
        UIDecimalSelector count(1001, 0, st.MAX_SHIPS - st.getNumShips(), 1);
        if (count.doStandardDialog(_("Replicate Ship"), _("Number:"), "pcc2:bsim"))
            replicateCurrentShip(count.getValue());
    }
}

/** Delete current ship. */
void
WSimListWithHandler::deleteShip()
{
    GSimState& st = getSimState();
    if (st.getNumShips() != 0) {
        st.removeShip(getCurrent());
    }
}

/** Write this ship back to game. */
void
WSimListWithHandler::writeToGame()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (game().getShipRelation(*sh) == GSimulatorGameInterface::ShipIsPlayed) {
            /* only try ships that are played; for others, the 'w'
               button is disabled */
            if (!game().updateToGame(*sh))
                messageBox(_("Ship status could not be written back."),
                           _("Battle Simulator"));
        }
    }
}

/** Update this ship from the game. */
void
WSimListWithHandler::updateFromGame()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (game().getShipRelation(*sh) != GSimulatorGameInterface::ShipIsUnknown) {
            /* only try ships that we recognize; for others, the 'u'
               button is disabled. In addition, updateFromGame() will
               update everything as long as the Id matches. */
            game().updateFromGame(*sh, WSimScreen::sim_opts);
        }
    }
}

/** Switch to game. */
void
WSimListWithHandler::switchToGame()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (game().selectGameShip(*sh)) {
            stop(1);
        }
    }
}

/** Edit aggressiveness. */
void
WSimListWithHandler::editAggressiveness()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIStringList list;
        list.add(GSimShip::agg_Kill, _("! - Kill Mission"));
        list.add(GSimShip::agg_Passive, _("0 - Passive Ship"));
        for (int i = 1; i <= NUM_OWNERS; ++i)
            list.add(i, format(_("%X - Primary Enemy %s"), i, player_racenames.getAdjName(i)));
        list.add(GSimShip::agg_NoFuel, _("Z - No Fuel"));

        UIStandardListbox box(0, list.getCount(), 20 * font_ems[FONT_NORMAL], list);
        box.scrollToKey(sh->getAggressiveness());
        if (box.doStandardListbox(_("Set Aggressiveness"), "pcc2:bsim")) {
            int new_val = box.getKey();
            sh->setAggressiveness(new_val);
            if (new_val == GSimShip::agg_Kill || new_val == GSimShip::agg_NoFuel) {
                sh->setFlags(sh->getFlags() & ~GSimShip::fl_Cloaked);
                sh->setInterceptId(0);
            }
        }
    }
}

/** Edit current ship's hull type.

    \param after_add true iff this is the invocation of this dialog
    immediately after adding a ship. In this case, also tries to set a
    default owner for the ship. */
void
WSimListWithHandler::editType(bool after_add)
{
    if (GSimShip* sh = getCurrentShip()) {
        int old_val = sh->getHull();
        int new_val = old_val;
        if (WHullSelectionDialog().run(_("Set Hull Type"), "pcc2:bsim", new_val)) {
            sh->setHull(new_val);

            if (new_val != old_val && new_val != 0 && after_add) {
                // pick default owner if possible
                int defo = 0;
                bool we_can_build_it = false;
                for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
                    if (getTruehullSlot(pl, new_val) != 0) {
                        if (defo == 0)
                            defo = pl;
                        else
                            defo = -1;
                        if (pl == sh->getOwner())
                            we_can_build_it = true;
                    }
                }
                if (defo > 0) {
                    if (defo != sh->getOwner()) {
                        /* ship has default owner */
                        sh->setOwner(defo);
                        sh->setFCode("?""?""?");
                        avoidSelfAgression(*sh);
                        limitDamage(*sh);
                    }
                } else {
                    if (!we_can_build_it) {
                        /* ship does not have a default owner, so we copy another ship's owner.
                           Do not copy the FCode, to avoid surprises */
                        sh->setFCode("?""?""?");
                    }
                }

                // Uncloak if it cannot cloak
                if (!sh->isCustomShip()
                    && !sh->hasImpliedSpecial(hf_Cloak)
                    && !sh->hasImpliedSpecial(hf_AdvancedCloak)
                    && !sh->hasImpliedSpecial(hf_HardenedCloak))
                {
                    sh->setFlags(sh->getFlags() & ~GSimShip::fl_Cloaked);
                }
            }
        }
    }
}

/** Edit current ship's friendly code. */
void
WSimListWithHandler::editFCode()
{
    if (GSimShip* sh = getCurrentShip()) {
        /* build list of available codes */
        GFCodeList filtered_list;
        GFCodeList& full_list = GFCodeList::getInstance();

        int our_flags = GFCode::fc_Ship;
        if (sh->getBayCount() != 0 || sh->getTorpLauncherCount() != 0 || sh->getBeamCount() != 0)
            our_flags |= GFCode::fc_Capital;

        for (GFCodeList::iterator i = full_list.begin(); i != full_list.end(); ++i) {
            if ((i->getFlags() & GFCode::fc_Ship) == 0)
                continue;
            if ((i->getFlags() & GFCode::fc_Capital) != 0 && (our_flags & GFCode::fc_Capital) == 0)
                continue;
            filtered_list.addFCode(*i);
        }

        WFCodeWindow window(_("Set Friendly Code"), filtered_list);
        window.setFCode(sh->getFCode());

        ui_root->centerWidget(window);
        if (ui_root->runModal(window) == cm_Enter) {
            sh->setFCode(window.getFCode());
            sh->assignRandomFCodeFlags();
        }
    }
}

/** Assign friendly code sequence. */
void
WSimListWithHandler::incrFCode()
{
    // Find current ship, scroll down, find new current ship
    GSimShip* prev = getCurrentShip();
    scrollTo(getCurrent()+1, 1);
    GSimShip* now = getCurrentShip();

    // If both exist and differ, perform our operation
    if (prev != 0 && now != 0 && prev != now) {
        // Copy friendly code and flags
        string_t fcode = prev->getFCode();
        int      flags = ((now->getFlags() & ~GSimShip::fl_RandomDigits)
                          | (prev->getFlags() & GSimShip::fl_RandomDigits));

        // Initialize fcode
        if (fcode.size() < 3)
            fcode.append(3 - fcode.size(), ' ');
        for (int i = 0; i < 3; ++i) {
            if (((flags & GSimShip::fl_RandomFC) != 0
                 && (flags & (GSimShip::fl_RandomFC1 << i)) != 0)
                || (fcode[i] < '0' || fcode[i] > '9'))
            {
                fcode[i] = '0' + (rand() % 10);
            }
        }

        // Find new fcode. Attempt to make it unique, but don't
        // crash if that's not possible (because we've made the fcode
        // completely numeric above, that will not happen anyway, but
        // it might happen if someone makes a .ccb file with >1000
        // entries).
        for (int tries = 0; tries < NUM_SHIPS; ++tries) {
            // Increase fcode
            for (int i = 3; i > 0; --i) {
                ++fcode[i-1];
                if (fcode[i-1] <= '9')
                    break;
                fcode[i-1] = '0';
            }

            // Check for uniqueness
            bool ok = true;
            for (int i = 0; i < int(getSimState().getNumShips()) && ok; ++i)
                if (i != getCurrent() && getSimState().getShip(i).getFCode() == fcode)
                    ok = false;
            if (ok)
                break;
        }

        // Assign new fcode
        now->setFCode(fcode);
        now->setFlags(flags);
    }
}

/** Edit current ship's primary weapon. */
void
WSimListWithHandler::editPrimaryWeapon()
{
    if (GSimShip* sh = getCurrentShip()) {
        int max_beams = 0;
        if (sh->isCustomShip()) {
            max_beams = 20;
        } else {
            max_beams = getHull(sh->getHull()).getMaxBeams();
        }
        if (sh->getBeamCount() != 0 || max_beams != 0) {
            /* build combo box */
            UIStringList list;
            list.add(0, _("No weapon"));
            for (int i = 1; i <= NUM_BEAMS; ++i)
                list.add(i, describeBeam(getBeam(i)));
            UIComboBox type_select(1000, 0, NUM_BEAMS, list);
            if (sh->getBeamCount() == 0)
                type_select.setValue(NUM_BEAMS);
            else
                type_select.setValue(sh->getBeamType());

            /* count */
            UIDecimalSelector count_select(2001, 0, max_beams, 1);
            count_select.setValue(sh->getBeamCount());

            /* build window */
            ObjectHolder h;
            UIWindow w(_("Set Primary Weapon"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
            w.setLayout(UIVBoxLayout::instance5);

            UIGroup& controls = h.add(new UIGroup(h.add(new UIGridLayout(2))));
            controls.add(h.add(new UIStaticText(_("Type:"), UIStaticText::StaticTitle)));
            createComboBoxControls(controls, type_select, h);
            controls.add(h.add(new UIStaticText(_("Count:"), UIStaticText::StaticTitle)));
            controls.add(wrapIncrement(h, 1001, count_select));

            UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
            buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
            buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
            buttons.add(h.add(new UISpacer()));

            w.add(controls);
            w.add(buttons);
            w.add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Vert)));
            w.add(h.add(new UIQuit(cm_Escape)));
            w.pack();
            type_select.setFocusedWidget();

            ui_root->centerWidget(w);
            int rv = ui_root->runModal(w);

            if (rv == cm_Enter) {
                if (type_select.getValue() != 0 && count_select.getValue() != 0) {
                    sh->setBeamCount(count_select.getValue());
                    sh->setBeamType(type_select.getValue());
                } else {
                    sh->setBeamCount(0);
                    sh->setBeamType(0);
                }
            }
        }
    }
}


/** Edit current ship's secondary weapon. */
void
WSimListWithHandler::editSecondaryWeapon()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (sh->isCustomShip()) {
            /* custom ship can have any combination */
            /* first build list of types */
            UIStringList list;
            list.add(0, _("No weapon"));
            for (int i = 1; i <= NUM_TORPS; ++i)
                list.add(i, describeTorpedo(getTorp(i)));
            list.add(NUM_TORPS+1, _("Fighters"));

            UIComboBox type_select(1000, 0, NUM_TORPS+1, list);

            /* count */
            UIDecimalSelector count_select(2001, 0, 20, 1);
            count_select.setValue(sh->getTorpLauncherCount());

            if (sh->getBayCount() != 0) {
                type_select.setValue(NUM_TORPS+1);
                count_select.setValue(sh->getBayCount());
            } else if (sh->getTorpLauncherCount() != 0) {
                type_select.setValue(sh->getTorpType());
                count_select.setValue(sh->getTorpLauncherCount());
            } else {
                type_select.setValue(NUM_TORPS);
                count_select.setValue(0);
            }

            /* ammo */
            UIDecimalSelector ammo_select(2002, 0, 10000, 10);
            ammo_select.setValue(sh->getAmmo());

            /* build window */
            ObjectHolder h;
            UIWindow w(_("Set Secondary Weapon"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
            w.setLayout(UIVBoxLayout::instance5);

            UIGroup& controls = h.add(new UIGroup(h.add(new UIGridLayout(2))));
            controls.add(h.add(new UIStaticText(_("Type:"), UIStaticText::StaticTitle)));
            createComboBoxControls(controls, type_select, h);
            controls.add(h.add(new UIStaticText(_("Count:"), UIStaticText::StaticTitle)));
            controls.add(wrapIncrement(h, 1001, count_select));
            controls.add(h.add(new UIStaticText(_("Ammo:"), UIStaticText::StaticTitle)));
            controls.add(wrapIncrement(h, 1002, ammo_select));

            UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
            buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
            buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
            buttons.add(h.add(new UISpacer()));

            w.add(controls);
            w.add(buttons);
            w.add(h.add(new UIFocusIterator(1000, 1002, UIFocusIterator::fi_Vert)));
            w.add(h.add(new UIQuit(cm_Escape)));
            w.pack();
            type_select.setFocusedWidget();

            ui_root->centerWidget(w);
            int rv = ui_root->runModal(w);

            if (rv == cm_Enter) {
                sh->setAmmo(ammo_select.getValue());
                if (count_select.getValue() != 0 && type_select.getValue() != 0) {
                    if (type_select.getValue() == NUM_TORPS+1) {
                        sh->setTorpLauncherCount(0);
                        sh->setTorpType(0);
                        sh->setBayCount(count_select.getValue());
                    } else {
                        sh->setTorpLauncherCount(count_select.getValue());
                        sh->setTorpType(type_select.getValue());
                        sh->setBayCount(0);
                    }
                } else {
                    sh->setTorpLauncherCount(0);
                    sh->setTorpType(0);
                    sh->setBayCount(0);
                }
            }
        } else {
            int max_tubes = getHull(sh->getHull()).getMaxLaunchers();
            int num_bays = getHull(sh->getHull()).getNumBays();
            int max_ammo = getHull(sh->getHull()).getCargo();
            if (num_bays != 0) {
                /* it's a carrier; just ask for number of fighters. */
                sh->setBayCount(num_bays);
                sh->setTorpType(0);
                sh->setTorpLauncherCount(0);

                UIDecimalSelector select(0, 0, max_ammo, 10);
                select.setValue(sh->getAmmo());
                if (select.doStandardDialog(_("Set Number of Fighters"), _("Fighters:"), "pcc2:bsim"))
                    sh->setAmmo(select.getValue());
            } else if (max_tubes != 0) {
                /* it's a torper, so we have to do a more complicated
                   dialog. */
                sh->setBayCount(0);

                /* first build list of types */
                UIStringList list;
                list.add(0, _("No weapon"));
                for (int i = 1; i <= NUM_TORPS; ++i) {
                    list.add(i, describeTorpedo(getTorp(i)));
                }

                UIComboBox type_select(1000, 0, NUM_TORPS, list);
                if (sh->getTorpLauncherCount() == 0)
                    type_select.setValue(NUM_TORPS);
                else
                    type_select.setValue(sh->getTorpType());

                /* count */
                UIDecimalSelector count_select(2001, 0, max_tubes, 1);
                count_select.setValue(sh->getTorpLauncherCount());

                /* ammo */
                UIDecimalSelector ammo_select(2002, 0, max_ammo, 10);
                ammo_select.setValue(sh->getAmmo());

                /* build window */
                ObjectHolder h;
                UIWindow w(_("Set Secondary Weapon"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
                w.setLayout(UIVBoxLayout::instance5);

                UIGroup& controls = h.add(new UIGroup(h.add(new UIGridLayout(2))));
                controls.add(h.add(new UIStaticText(_("Type:"), UIStaticText::StaticTitle)));
                createComboBoxControls(controls, type_select, h);
                controls.add(h.add(new UIStaticText(_("Count:"), UIStaticText::StaticTitle)));
                controls.add(wrapIncrement(h, 1001, count_select));
                controls.add(h.add(new UIStaticText(_("Torpedoes:"), UIStaticText::StaticTitle)));
                controls.add(wrapIncrement(h, 1002, ammo_select));

                UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
                buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
                buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
                buttons.add(h.add(new UISpacer()));

                w.add(controls);
                w.add(buttons);
                w.add(h.add(new UIFocusIterator(1000, 1002, UIFocusIterator::fi_Vert)));
                w.add(h.add(new UIQuit(cm_Escape)));
                w.pack();
                type_select.setFocusedWidget();

                ui_root->centerWidget(w);
                int rv = ui_root->runModal(w);

                if (rv == cm_Enter) {
                    sh->setAmmo(ammo_select.getValue());
                    if (count_select.getValue() != 0 && type_select.getValue() != 0) {
                        sh->setTorpLauncherCount(count_select.getValue());
                        sh->setTorpType(type_select.getValue());
                    } else {
                        sh->setTorpLauncherCount(0);
                        sh->setTorpType(0);
                    }
                }
            }
        }
    }
}

/** Edit current ship's name. */
void
WSimListWithHandler::editName()
{
    if (GSimShip* sh = getCurrentShip()) {
        string_t s = sh->getName();
        if (UIInputLine::run(s, 20, _("Set Ship Name"), _("Name:"), 20, UIInputLine::ilf_GameChars))
            sh->setName(s);
    }
}

/** Edit current ship's Id number. */
void
WSimListWithHandler::editId()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIDecimalSelector select(0, 1, NUM_SHIPS, 10);
        select.setValue(sh->getId());
        while (1) {
            if (!select.doStandardDialog(_("Set Id"), _("Id:"), "pcc2:bsim"))
                break;

            bool id_is_in_use = false;
            for (GSimState::ship_index_t i = 0; i < getSimState().getNumShips(); ++i) {
                GSimShip& ish = getSimState().getShip(i);
                if (&ish != sh && ish.getId() == select.getValue()) {
                    id_is_in_use = true;
                    break;
                }
            }

            if (!id_is_in_use) {
                bool isDefault = sh->getName() == makeDefaultName(sh->getId());
                sh->setId(select.getValue());
                if (isDefault) {
                    sh->setName(makeDefaultName(sh->getId()));
                }
                break;
            }

            messageBox(_("This Id number is already in use in this simulation setup. Please choose another one."),
                       _("Battle Simulator"));
        }
    }
}

/** Edit current ship's damage level. */
void
WSimListWithHandler::editDamage()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIDecimalSelector select(0, 0, sh->getMaxDamage(), 10);
        select.setValue(sh->getDamage());
        if (select.doStandardDialog(_("Set Damage"), _("Damage [%]:"), "pcc2:bsim")) {
            int val = select.getValue();
            sh->setDamage(val);
            if (sh->getShield() > sh->getMaxShield())
                sh->setShield(sh->getMaxShield());
        }
    }
}

/** Edit current ship's shield level. */
void
WSimListWithHandler::editShield()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIDecimalSelector select(0, 0, sh->getMaxShield(), 10);
        select.setValue(sh->getShield());
        if (select.doStandardDialog(_("Set Shield Level"), _("Shield [%]:"), "pcc2:bsim"))
            sh->setShield(select.getValue());
    }
}

/** Edit current ship's mass. This only works on custom ships. */
void
WSimListWithHandler::editMass()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (sh->isCustomShip()) {
            UIDecimalSelector select(0, 1, 10000, 100);
            select.setValue(sh->getMass());
            if (select.doStandardDialog(_("Set Mass"), _("Mass [kt]:"), "pcc2:bsim"))
                sh->setMass(select.getValue());
        }
    }
}

/** Edit current ship's crew amount. */
void
WSimListWithHandler::editCrew()
{
    if (GSimShip* sh = getCurrentShip()) {
        int limit;
        if (sh->isCustomShip())
            limit = 10000;
        else
            limit = getHull(sh->getHull()).getCrew();

        // All ships are allowed to have 10 crew, because host sometimes sets it so.
        if (limit < 10) {
            limit = 10;
        }

        UIDecimalSelector select(0, 1, limit, 10);
        select.setValue(sh->getCrew());
        if (select.doStandardDialog(_("Set Crew"), _("Crew:"), "pcc2:bsim"))
            sh->setCrew(select.getValue());
    }
}

/** Edit current ship's experience level. */
void
WSimListWithHandler::editExperienceLevel()
{
    if (GSimShip* sh = getCurrentShip())
        ::editExperienceLevel(*sh);
}

/** Edit current ship's owner. */
void
WSimListWithHandler::editOwner()
{
    if (GSimShip* sh = getCurrentShip()) {
        ::editOwner(*sh);
        avoidSelfAgression(*sh);
    }
}

/** Edit current ship's engine type. */
void
WSimListWithHandler::editEngine()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIStringList list;
        for (int i = 1; i <= NUM_ENGINES; ++i)
            list.add(i, format("%X - %s", i, getEngine(i).getName()));

        UIStandardListbox box(0, list.getCount(), font_ems[FONT_NORMAL] * 20, list);
        box.scrollToKey(sh->getEngineType());
        if (box.doStandardListbox(_("Set Engine Type"), "pcc2:bsim"))
            sh->setEngineType(box.getKey());
    }
}

/** Edit current ship's intercept target. */
void
WSimListWithHandler::editInterceptId()
{
    if (GSimShip* sh = getCurrentShip()) {
        /* FIXME: use a better widget */
        UIDecimalSelector select(0, 0, NUM_SHIPS, 10);
        select.setValue(sh->getInterceptId());
        if (select.doStandardDialog(_("Set Intercept-Attack Target"), _("Id:"), "pcc2:bsim")) {
            sh->setInterceptId(select.getValue());
            if (sh->getInterceptId() != 0) {
                /* cannot cloak and intercept */
                sh->setFlags(sh->getFlags() & ~GSimShip::fl_Cloaked);
                if (const GSimShip* target = getSimState().getShipById(sh->getInterceptId())) {
                    /* pick correct PE */
                    sh->setAggressiveness(target->getOwner());
                } else if (int n = game().getShipOwner(sh->getInterceptId())) {
                    /* we're intercepting a ship which is not in the
                       simulation, but in the game. Set correct PE in
                       case ship is added later */
                    sh->setAggressiveness(n);
                } else if (sh->getAggressiveness() == GSimShip::agg_Kill || sh->getAggressiveness() == GSimShip::agg_NoFuel) {
                    /* cannot intercept and kill, cannot intercept
                       without fuel. */
                    sh->setAggressiveness(0);
                }
            }
        }
    }
}

/** Toggle cloak status on current ship. */
void
WSimListWithHandler::toggleCloak()
{
    if (GSimShip* sh = getCurrentShip()) {
        int flags = sh->getFlags();
        flags ^= GSimShip::fl_Cloaked;
        if (flags & GSimShip::fl_Cloaked) {
            /* cannot cloak and be aggressive */
            if (sh->getAggressiveness() == GSimShip::agg_Kill || sh->getAggressiveness() == GSimShip::agg_NoFuel)
                sh->setAggressiveness(GSimShip::agg_Passive);
            /* cannot cloak and intercept */
            sh->setInterceptId(0);
        }
        sh->setFlags(flags);
    }
}

/** Toggle "random FCode" function. */
void
WSimListWithHandler::toggleRandomFC()
{
    if (GSimShip* sh = getCurrentShip()) {
        sh->setFlags(sh->getFlags() ^ GSimShip::fl_RandomFC);
    }
}

/** Toggle whether thip is enabled. */
void
WSimListWithHandler::toggleEnabled()
{
    if (GSimShip* sh = getCurrentShip()) {
        sh->setFlags(sh->getFlags() ^ GSimShip::fl_Deactivated);
    }
}

/** Replicate current ship. This is the back-end function. Caller must
    make sure that the maximum possible number of ships is not
    exceeded if he wants to notify user. This function will silently
    ignore failures. Failure here means inability to obtain a free Id
    number, which only happens when there are more than NUM_SHIPS
    ships.

    \param count number of times to replicate */
void
WSimListWithHandler::replicateCurrentShip(int count)
{
    GSimState& st = getSimState();

    // Build bitset of used ship Ids. The naive version of this routine,
    // using getFreeId() to obtain new ship Ids, is O(n^3), taking a
    // noticeable amount of time (>1s at 2 GHz) when inserting 998 ships
    // (=the maximum).
    uint8 used_id_bitset[NUM_SHIPS/8 + 1];
    zeroFill(used_id_bitset);
    for (GSimState::ship_index_t i = 0; i < st.getNumShips(); ++i) {
        int id = st.getShip(i).getId();
        if (id > 0 && id <= NUM_SHIPS)
            used_id_bitset[id >> 3] |= (1 << (id & 7));
    }

    // Go
    int pos = getCurrent();
    int id = 0;
    for (int i = 0; i < count; ++i) {
        do
            ++id;
        while (id <= NUM_SHIPS && (used_id_bitset[id >> 3] & (1 << (id & 7))) != 0);
        if (id > NUM_SHIPS)
            break;
        used_id_bitset[id >> 3] |= (1 << (id & 7));
        st.duplicateShip(pos, id);
        ++pos;
    }
    getSimState().doUpdates();
    scrollTo(pos, 1);
}

/**************************** WSimPlanetButton ***************************/

/** Constructor.
    \param state State to observe */
WSimPlanetButton::WSimPlanetButton(GSimulatorGameInterface& game_if, GSimState& state)
    : UIAbstractButton(UIKey('p')),
      state(state),
      game_if(game_if),
      conn_edit(observe(state.sig_planet_changed)),
      conn_structure(observe(state.sig_structure_changed))
{ }

WSimPlanetButton::~WSimPlanetButton()
{ }

void
WSimPlanetButton::drawContent(GfxCanvas& can)
{
    GfxRect r = getExtent();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    drawFrameDown(can, r);
    r.grow(-1, -1);
    drawSolidBar(ctx, r, COLOR_GRAY);
    r.grow(-4, -3);

    ctx.useFont(*fonts[FONT_NORMAL]);
    if (state.hasPlanet()) {
        ctx.setColor(COLOR_BLACK);
        outTextF(ctx, r.x, r.y, r.w, getPlanetName(game_if, state));
    } else {
        ctx.setColor(COLOR_DARK);
        outTextF(ctx, r.x, r.y, r.w, _("(no planet or base)"));
    }
    can.update(getExtent());
}

void
WSimPlanetButton::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = font_sizes[FONT_NORMAL].scaledBy(10, 1).movedBy(0, 10);
    info.pref_size = font_sizes[FONT_NORMAL].scaledBy(18, 1).movedBy(0, 10);
    info.flags = li_FlexibleH;
}

/********************************* Tiles *********************************/

/** Simulator ship information tile.

    Displays information about the current ship in a simulation. This
    class provides the necessary interface to WSimList; descendants
    only need to override the display methods. */
class WSimShipTile : public WTile {
 protected:
    WSimList& list;
 private:
    SignalConnection conn_moved, conn_ship_changed, conn_structure_changed;

 protected:
    WSimShipTile(WSimList& list, int width, int height);
    ~WSimShipTile();

    void onChange();
    void onShipEdit(int slot);
};

WSimShipTile::WSimShipTile(WSimList& list, int width, int height)
    : WTile(GfxPoint(width, height)), list(list),
      conn_moved(list.sig_moved.add(this, &WSimShipTile::onChange)),
      conn_ship_changed(list.getSimState().sig_ship_changed.add(this, &WSimShipTile::onShipEdit)),
      conn_structure_changed(list.getSimState().sig_structure_changed.add(this, &WSimShipTile::onChange))
{ }

WSimShipTile::~WSimShipTile()
{ }

void
WSimShipTile::onShipEdit(int slot)
{
    if (slot == list.getCurrent())
        drawWidget();
}

void
WSimShipTile::onChange()
{
    drawWidget();
}

/**************************** WSimShipTypeTile ***************************/

class WSimShipTypeTile : public WSimShipTile {
    ObjectHolder h;
    void init(GSimulatorGameInterface& game_if, WSimList& list);
 public:
    WSimShipTypeTile(GSimulatorGameInterface& game_if, WSimList& list);
    void drawData(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
};

class WSimShipDamageTile : public WSimShipTile {
    void drawContent(UITableWriter& tw, bool with_data);
    ObjectHolder h;
    void init(GSimulatorGameInterface& game_if, WSimList& list);
 public:
    WSimShipDamageTile(GSimulatorGameInterface& game_if, WSimList& list);
    void drawData(GfxCanvas& can);
};

class WSimShipWeaponTile : public WSimShipTile {
    void drawContent(UITableWriter& tw, bool with_data);
    ObjectHolder h;
    void init(GSimulatorGameInterface& game_if, WSimList& list);
 public:
    WSimShipWeaponTile(GSimulatorGameInterface& game_if, WSimList& list);
    void drawData(GfxCanvas& can);
};

class WSimShipImageTile : public WSimShipTile {
 public:
    WSimShipImageTile(WSimList& list);
    void drawData(GfxCanvas& can);
};

/** Button on ship editor. Automatically manages its disabledness
    state depending on current ship. */
class WSimShipButton : public UIButton {
    WSimList& list;
    GSimulatorGameInterface& game_if;
    SignalConnection conn_moved, conn_ship_changed, conn_structure_changed;

 public:
    WSimShipButton(GSimulatorGameInterface& game_if, WSimList& list, const char* name, UIKey key);
    ~WSimShipButton();

    void onUpdate();
};

WSimShipButton::WSimShipButton(GSimulatorGameInterface& game_if, WSimList& list, const char* name, UIKey key)
    : UIButton(name, key),
      list(list),
      game_if(game_if),
      conn_moved(list.sig_moved.add(this, &WSimShipButton::onUpdate)),
      conn_ship_changed(list.getSimState().sig_ship_changed.add(this, &WSimShipButton::onUpdate)),
      conn_structure_changed(list.getSimState().sig_structure_changed.add(this, &WSimShipButton::onUpdate))
{
    if (name[0] != 0 && name[1] == 0 && name[0] != 'U' && name[0] != 'W')
        setState(bf_SmallSize, true);
    onUpdate();
}

WSimShipButton::~WSimShipButton()
{ }

void
WSimShipButton::onUpdate()
{
    bool enable;
    bool press = false;
    const GSimState& st = list.getSimState();
    if (list.getCurrent() >= int(st.getNumShips())) {
        /* no ships => all controls disabled */
        enable = false;
    } else {
        const GSimShip& this_ship = list.getSimState().getShip(list.getCurrent());
        switch (getKey()) {
         case SDLK_RETURN:
            enable = (st.getNumShips() > 1 || st.hasPlanet());
            break;

         case '1':
            enable = (this_ship.isCustomShip()
                      || getHull(this_ship.getHull()).getMaxBeams() != 0);
            break;

         case '2':
            enable = (this_ship.isCustomShip()
                      || getHull(this_ship.getHull()).getMaxLaunchers() != 0
                      || getHull(this_ship.getHull()).getNumBays() != 0);
            break;

         case SDLK_F1:
         case 'w':
            enable = (game_if.getShipRelation(this_ship) == GSimulatorGameInterface::ShipIsPlayed);
            break;

         case 'u':
            enable = (game_if.getShipRelation(this_ship) != GSimulatorGameInterface::ShipIsUnknown);
            break;

         case 'm':
            enable = this_ship.isCustomShip();
            break;

         case 'l':
            enable = WSimScreen::isExpGame();
            break;

         default:
            enable = true;
            break;
        }

        if (getKey() == 'r') {
            press = (this_ship.getFlags() & GSimShip::fl_RandomFC) != 0;
        }
    }

    setState(st_Disabled, !enable);
    setState(bf_Highlight, press);
}

WSimShipTypeTile::WSimShipTypeTile(GSimulatorGameInterface& game_if, WSimList& list)
    : WSimShipTile(list, 15 * font_ems[FONT_NORMAL], 6 * font_heights[FONT_NORMAL])
{
    init(game_if, list);
}

void
WSimShipTypeTile::init(GSimulatorGameInterface& game_if, WSimList& list)
{
    // Translators: this button always is the 'r' key.
    const char* r = _("Random#");
    int em = font_heights[FONT_NORMAL];
    int em1 = em-1;
    int rwi = fonts[FONT_SMALL]->getTextWidth(r) + 8;
    int rwi5 = rwi+2;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(rwi5,    0, em1, em1), h.add(new WSimShipButton(game_if, list, "N", 'n')));
    addControl(GfxRect(rwi5,   em, em1, em1), h.add(new WSimShipButton(game_if, list, "T", 't')));
    addControl(GfxRect(rwi5, 2*em, em1, em1), h.add(new WSimShipButton(game_if, list, "O", 'o')));
    addControl(GfxRect(rwi5, 3*em, em1, em1), h.add(new WSimShipButton(game_if, list, "I", 'i')));
    addControl(GfxRect(rwi5, 4*em, em1, em1), h.add(new WSimShipButton(game_if, list, "F", 'f')));

    WSimShipButton& rb = h.add(new WSimShipButton(game_if, list, r, 'r'));
    rb.setState(rb.bf_SmallSize, true);
    addControl(GfxRect(0,    4*em, rwi, em1), rb);
}

void
WSimShipTypeTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(0, font_heights[FONT_NORMAL]), FONT_NORMAL);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Name:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Type:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Owner:"));
    tw.writeFixedCell(UIColor::tc_Static, _("ID:"));
    tw.writeFixedCell(UIColor::tc_Static, _("FCode:"));

    tw.newLColumnUpTo(getExtent().w);
    if (const GSimShip* sh = list.getCurrentShip()) {
        tw.writeVariableCell(UIColor::tc_Green, sh->getName());
        if (sh->isCustomShip())
            tw.writeVariableCell(UIColor::tc_Green, _("Custom Ship"));
        else
            tw.writeVariableCell(UIColor::tc_Green, getHull(sh->getHull()).getName());
        tw.writeVariableCell(UIColor::tc_Green, player_racenames.getShortName(sh->getOwner()));
        tw.writeVariableCell(UIColor::tc_Green, itoa(sh->getId()));

        // Friendly-code is special: use color-coding for randomized
        // digits. This cannot be done directly with UITableWriter.
        // Hence, remember current position and write a blank cell,
        // then write fcode manually.
        GfxPoint pos = tw.getCurrentPosition();
        tw.writeVariableCell(UIColor::tc_Green, string_t());
        drawRandomFCode(ctx, pos, *sh);
    } else {
        for (int i = 0; i < 5; ++i)
            tw.writeVariableCell(UIColor::tc_Green, string_t());
    }
    tw.finish();
}

void
WSimShipTypeTile::getLayoutInfo(LayoutInfo& info)
{
    /* This tile sits on the same line as the WSimShipImageTile, and we want it
       to occupy as many horizontal space as possible. The WSimShipImageTile
       should not be scaled. */
    WSimShipTile::getLayoutInfo(info);
    info.flags |= li_FlexibleH;
}

WSimShipDamageTile::WSimShipDamageTile(GSimulatorGameInterface& game_if, WSimList& list)
    : WSimShipTile(list, 20 * font_ems[FONT_NORMAL], 3 * font_heights[FONT_NORMAL])
{
    init(game_if, list);
}

void
WSimShipDamageTile::init(GSimulatorGameInterface& game_if, WSimList& list)
{
    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(em,     0, em1, em1), h.add(new WSimShipButton(game_if, list, "D", 'd')));
    addControl(GfxRect(0,     em, em1, em1), h.add(new WSimShipButton(game_if, list, "L", 'l')));
    addControl(GfxRect(em,    em, em1, em1), h.add(new WSimShipButton(game_if, list, "C", 'c')));
    addControl(GfxRect(0,   2*em, em1, em1), h.add(new WSimShipButton(game_if, list, "M", 'm')));
    addControl(GfxRect(em,  2*em, em1, em1), h.add(new WSimShipButton(game_if, list, "S", 's')));
}

void
WSimShipDamageTile::drawContent(UITableWriter& tw, bool with_data)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Crew:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Shield:"));

    if (with_data) {
        if (const GSimShip* sh = list.getCurrentShip()) {
            tw.newRColumn(5 * font_ems[FONT_NORMAL]);
            tw.writeVariableCell(UIColor::tc_Green, numToString(sh->getDamage()));
            tw.writeVariableCell(UIColor::tc_Green, numToString(sh->getCrew()));
            tw.writeVariableCell(UIColor::tc_Green, numToString(sh->getShield()));

            tw.newLColumnUpTo(2 * font_ems[FONT_NORMAL]);
            tw.writeVariableCell(UIColor::tc_Green, "%");
            tw.writeVariableCell(UIColor::tc_Green, "");
            tw.writeVariableCell(UIColor::tc_Green, "%");

            tw.newLColumn();
            tw.writeVariableCell(UIColor::tc_Green, "");
            if (WSimScreen::isExpGame())
                tw.writeVariableCell(UIColor::tc_Green, getExperienceLevelName(sh->getExperienceLevel()));
            else
                tw.writeVariableCell(UIColor::tc_Green, "");
            int mass = sh->getMass();
            tw.writeVariableCell(UIColor::tc_Green, format(_("%d kt"), numToString(mass)));
        }
    }
    tw.finish();
}

void
WSimShipDamageTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(0, font_heights[FONT_NORMAL]*2), FONT_NORMAL);
    drawContent(tw, true);
}



WSimShipWeaponTile::WSimShipWeaponTile(GSimulatorGameInterface& game_if, WSimList& list)
    : WSimShipTile(list, 25 * font_ems[FONT_NORMAL], 8 * font_heights[FONT_NORMAL])
{
    init(game_if, list);
}

void
WSimShipWeaponTile::init(GSimulatorGameInterface& game_if, WSimList& list)
{
    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(em,         0, em1, em1), h.add(new WSimShipButton(game_if, list, "1", '1')));
    addControl(GfxRect(em,        em, em1, em1), h.add(new WSimShipButton(game_if, list, "2", '2')));
    addControl(GfxRect(em,      3*em, em1, em1), h.add(new WSimShipButton(game_if, list, "E", 'e')));
    addControl(GfxRect(em,      4*em, em1, em1), h.add(new WSimShipButton(game_if, list, "A", 'a')));
    addControl(GfxRect(0,       4*em, em1, em1), h.add(new WSimShipButton(game_if, list, "V", 'v')));
    addControl(GfxRect(em,      5*em, em1, em1), h.add(new WSimShipButton(game_if, list, "B", 'b')));
    addControl(GfxRect(em,      6*em, em1, em1), h.add(new WSimShipButton(game_if, list, "X", 'x')));
    addControl(GfxRect(em,      7*em, em1, em1), h.add(new WSimShipButton(game_if, list, "K", 'k')));
}

void
WSimShipWeaponTile::drawContent(UITableWriter& tw, bool with_data)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Primary:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Secondary:"));
    tw.writeFixedCell(UIColor::tc_Static, "");
    tw.writeFixedCell(UIColor::tc_Static, _("Engine:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Aggressive:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Abilities:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Intercept-Att.:"));
    tw.writeFixedCell(UIColor::tc_Static, _("FLAK Rating:"));

    if (with_data) {
        if (const GSimShip* sh = list.getCurrentShip()) {
            tw.newLColumn();
            if (sh->getBeamCount() != 0)
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d \xC3\x97 %s"), sh->getBeamCount(), getBeam(sh->getBeamType()).getName()));
            else
                tw.writeVariableCell(UIColor::tc_Green, _("none"));

            if (sh->getTorpLauncherCount() != 0) {
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d %s launcher%!1{s%}"), sh->getTorpLauncherCount(), getTorp(sh->getTorpType()).getName()));
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d torpedo%!1{es%}"), numToString(sh->getAmmo())));
            } else if (sh->getBayCount() != 0) {
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d fighter bay%!1{s%}"), sh->getBayCount()));
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d fighter%!1{s%}"), numToString(sh->getAmmo())));
            } else {
                tw.writeVariableCell(UIColor::tc_Green, _("none"));
                tw.writeVariableCell(UIColor::tc_Green, "");
            }

            tw.writeVariableCell(UIColor::tc_Green, getEngine(sh->getEngineType()).getName());

            string_t agg;
            if (sh->getAggressiveness() == sh->agg_Kill)
                agg = _("Kill");
            else if (sh->getAggressiveness() >= 1 && sh->getAggressiveness() <= NUM_OWNERS)
                agg = format(_("Enemy %s"), player_racenames.getAdjName(sh->getAggressiveness()));
            else if (sh->getAggressiveness() == sh->agg_NoFuel)
                agg = _("No fuel");
            else
                agg = _("Passive");
            if (sh->getFlags() & sh->fl_Cloaked)
                agg += _(", cloaked");
            tw.writeVariableCell(UIColor::tc_Green, agg);

            agg = describeFunctions(*sh, WSimScreen::sim_opts);
            if (sh->hasAnyNonstandardFunction())
                tw.writeVariableCell(UIColor::tc_Yellow, agg);
            else
                tw.writeVariableCell(UIColor::tc_Green, agg);

            if (sh->getInterceptId() == 0)
                agg = _("none");
            else if (const GSimShip* other = list.getSimState().getShipById(sh->getInterceptId()))
                agg = format("%s (#%d)", other->getName(), other->getId());
            else
                agg = format(_("Ship #%d"), sh->getInterceptId());
            tw.writeVariableCell(UIColor::tc_Green, agg);

            if ((sh->getFlags() & GSimShip::fl_RatingOverride) == 0) {
                tw.writeVariableCell(UIColor::tc_Green, _("default"));
            } else {
                tw.writeVariableCell(UIColor::tc_Green, format("%d / %d", sh->getFlakRatingOverride(), sh->getFlakCompensationOverride()));
            }
        }
    }
    tw.finish();
}

void
WSimShipWeaponTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(0, font_heights[FONT_NORMAL]*2), FONT_NORMAL);
    drawContent(tw, true);
}


WSimShipImageTile::WSimShipImageTile(WSimList& list)
    : WSimShipTile(list, 105+2, 93+2)
{ }

void
WSimShipImageTile::drawData(GfxCanvas& can)
{
    GfxRect inner = getInnerSize(0, 0);
    drawSolidBar(can, inner, standard_colors[0]);
    drawFrameDown(can, inner);

    if (GSimShip* sh = list.getCurrentShip()) {
        const int hullnr = sh->getHull();

        Ptr<GfxPixmap> pix;
        if (hullnr > 0)
            pix = requestPixmap(ResId(res::ship, getHull(hullnr).getInternalPictureNumber(), hullnr));
        else
            pix = requestPixmap(ResId(res::ship, 200, 0));

        if (pix) {
            GfxRect clip(0, 0,
                         std::min(inner.w - 2, pix->getWidth()),
                         std::min(inner.h - 2, pix->getHeight()));
            can.blitSurface(GfxPoint(1, 1), *pix, clip);
        }
    } else {
        // TRANSLATIONS: Must be two lines
        string_t text = _("Press [Ins] to\nadd a ship");

        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.useFont(*fonts[FONT_SMALL]);
        ctx.setTextAlign(1, 1);
        ctx.setColor(COLOR_WHITE);

        GfxPoint pt = inner.getCenter();
        outTextFormatted(ctx, pt.x, pt.y-font_heights[FONT_SMALL]/2, extent.w, text);
    }
}

/** Edit current ship's abilities. */
void
WSimListWithHandler::editAbilities()
{
    if (GSimShip* sh = getCurrentShip()) {
        editSimulatorShipFunctions(*sh, WSimScreen::sim_opts);
    }
}

/** Move current ship upwards by swapping with the ship above it. */
void
WSimListWithHandler::swapUp()
{
    int n = getCurrent();
    if (n > 0) {
        scrollTo(n-1, 1);
        getSimState().swapShips(n, n-1);
    }
}

/** Move current ship downwards by swapping with the ship below it. */
void
WSimListWithHandler::swapDown()
{
    int n = getCurrent();
    if (n + 1 < int(getSimState().getNumShips())) {
        scrollTo(n+1, 1);
        getSimState().swapShips(n, n+1);
    }
}

/******************************* WSimScreen ******************************/

GSimState WSimScreen::sim_state;
GSimOptions WSimScreen::sim_opts;
bool WSimScreen::sim_default_relations = true;

WSimScreen::WSimScreen(GSimulatorGameInterface& game_if)
    : UIWindow(_("Battle Simulator"), BLUE_WINDOW, COLOR_DARK),
      ed_list(h.add(new WSimListWithHandler(game_if, sim_state))),
      game_if(game_if),
      btn_extras(h.add(new UIButton(0, _("F10 - Extras"), SDLK_F10)))
{
    init();
}

WSimScreen::~WSimScreen()
{ }

void
WSimScreen::init()
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& all_main             = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& all_buttons_1        = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& all_buttons_2        = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& all_main_list        = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& all_main_list_button = h.add(new UIGroup(UIHBoxLayout::instance0));
    UIFrameGroup& all_panels      = h.add(new UIFrameGroup(UIVBoxLayout::instance5, 5));
    UIGroup& all_panels_group     = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& ship_info_group      = h.add(new UIGroup(UIHBoxLayout::instance5));

    WSimPlanetButton& ed_planet = h.add(new WSimPlanetButton(game_if, sim_state));
    UIFrameGroup& ed_list_frame = h.add(new UIFrameGroup(UIVBoxLayout::instance0));
    ed_list_frame.add(ed_list);

    WSimShipWeaponTile& ssw_tile = h.add(new WSimShipWeaponTile(game_if, ed_list));
    ssw_tile.setHeading(_("Weapons"));
    ssw_tile.setDisplayState(WTile::sNormal);

    WSimShipTypeTile&   sst_tile = h.add(new WSimShipTypeTile(game_if, ed_list));
    WSimShipDamageTile& ssd_tile = h.add(new WSimShipDamageTile(game_if, ed_list));
    WSimShipImageTile&  ssi_tile = h.add(new WSimShipImageTile(ed_list));

    all_main_list_button.add(ed_planet);
    all_main_list_button.add(h.add(new UIButton("P", 'p')));

    all_main_list.add(h.add(new UIStaticText(_("Ships in simulation:"), UIStaticText::WhiteTitle)));
    all_main_list.add(ed_list_frame);
    all_main_list.add(all_main_list_button);

    ship_info_group.add(ssi_tile);
    ship_info_group.add(sst_tile);

    all_panels.add(ship_info_group);
    all_panels.add(ssd_tile);
    all_panels.add(ssw_tile);
    all_panels.add(h.add(new UISpacer()));
    if (game_if.isGameAvailable()) {
        UIGroup& game_group = h.add(new UIGroup(UIHBoxLayout::instance5));
        game_group.add(h.add(new WSimShipButton(game_if, ed_list, "U", 'u')));
        game_group.add(h.add(new WSimShipButton(game_if, ed_list, "W", 'w')));
        game_group.add(h.add(new UIStaticText(_("Update/Write back"), UIStaticText::WhiteTitle)));
        game_group.add(h.add(new UISpacer()));
        game_group.add(h.add(new WSimShipButton(game_if, ed_list, "F1", SDLK_F1)));
        all_panels.add(game_group);
    }

    all_panels_group.add(h.add(new UIStaticText(_("This ship:"), UIStaticText::WhiteTitle)));
    all_panels_group.add(all_panels);

    all_main.add(all_main_list);
    all_main.add(all_panels_group);

    all_buttons_1.add(h.add(new UIButton(_("Ins - Add"), SDLK_INSERT)));
    all_buttons_1.add(h.add(new WSimShipButton(game_if, ed_list, _("Del - Remove"), SDLK_DELETE)));
    all_buttons_1.add(h.add(new WSimShipButton(game_if, ed_list, _("Ctrl-S - Save"), ss_Ctrl + 's')));
    all_buttons_1.add(h.add(new UISpacer()));
    all_buttons_1.add(btn_extras);

    all_buttons_2.add(h.add(new WSimShipButton(game_if, ed_list, _("ENTER - Simulate"), SDLK_RETURN)));
    all_buttons_2.add(h.add(new UIButton(0, _("ESC"), SDLK_ESCAPE)).addStop());
    all_buttons_2.add(h.add(new UIButton(_("Ctrl-R - Load"), ss_Ctrl + 'r')));
    all_buttons_2.add(h.add(new UISpacer()));
    all_buttons_2.add(h.add(new UIButton(_("Help"), 'h')));

    add(all_main);
    add(all_buttons_1);
    add(all_buttons_2);
    add(h.add(new WHelpWidget("pcc2:bsim")));
    add(h.add(new UIQuit(0)));

    pack();

    ed_list.sig_menu.add(this, &WSimScreen::onContextMenu);
    btn_extras.sig_fired.add(this, &WSimScreen::onExtraMenu);

    if (sim_default_relations) {
        setDefaultRelations(game_if);
    }
}

/** Planet Editor. */
void
WSimScreen::editPlanet()
{
    editSimulatorPlanet(game_if, sim_state);
}

/** Load file. Ask for file name, then load the file and replace
    current selection with it. */
void
WSimScreen::loadFile()
{
    UIFileChooserDialog select(_("Load Simulation"),
                               Directory::openDir(Directory::getWorkingDirectory()),
                               FileNamePattern::getAllFilesWithExtensionPattern("ccb"));
    ui_root->centerWidget(select);
    bool rv = ui_root->runModal(select);

    if (rv) {
        string_t result = appendFileNameExtension(select.getResultFile(), "ccb", false);
        try {
            {
                Ptr<Stream> file = Stream::openFile(result, Stream::C_READ);
                GSimState tmp;
                tmp.load(*file);
                sim_state.swap(tmp);
            }
            ed_list.scrollTo(0, 1);
            if (sim_state.hasPlanet())
                setPlanetName(game_if, sim_state.getPlanet());
            sim_state.doUpdates();
            if (!sim_state.isMatchingShipList())
                messageBox(_("This simulation seems to have been set up with a different ship list. "
                             "To regenerate the intended results, restart the Simulator from the correct "
                             "game directory."),
                           _("Load Simulation"));
        }
        catch (FileFormatException& e) {
            messageBox(format(_("Unable to load %s: %s"), result, e.what()),
                       _("Load Simulation"));
        }
        catch (std::runtime_error& e) {
            messageBox(format(_("Unable to open %s: %s"), result, e.what()),
                       _("Load Simulation"));
        }
    }
}

/** Save file. */
void
WSimScreen::saveFile()
{
    UIFileChooserDialog select(_("Save Simulation"),
                               Directory::openDir(Directory::getWorkingDirectory()),
                               FileNamePattern::getAllFilesWithExtensionPattern("ccb"));
    ui_root->centerWidget(select);
    bool rv = ui_root->runModal(select);

    if (rv) {
        string_t result = appendFileNameExtension(select.getResultFile(), "ccb", false);
        try {
            Ptr<Stream> file = Stream::openFile(result, Stream::C_CREATE | Stream::C_WRITE);
            sim_state.save(*file);
        }
        catch (std::runtime_error& e) {
            messageBox(format(_("Unable to save %s: %s"), result, e.what()),
                       _("Save Simulation"));
        }
    }
}

/** Fleet cost comparison dialog. */
void
WSimScreen::doFleetCostComparison()
{
    if (sim_state.getNumShips() > 0) {
        WFleetCostDialog dlg(sim_state);
        ui_root->centerWidget(dlg);
        ui_root->runModal(dlg);
    }
}

/** Run the simulation (ENTER). Does the first simulation and opens
    the result window which allows users to run more simulations. */
void
WSimScreen::runSimulation()
{
    /* do we allow to run the simulation? */
    if (sim_state.getNumShips() == 0)
        return;
    if (sim_state.getNumShips() == 1 && !sim_state.hasPlanet())
        return;

    /* yes, we do. First randomize fcodes as requested. */
    if (!sim_opts.randomize_fcodes_on_every_fight) {
        sim_state.assignRandomFCodes();
        sim_state.doUpdates();
    }

    /* now try to get some results */
    WSimResultWindow result_window(game_if, sim_state, sim_opts);
    if (!result_window.runFirstSimulation()) {
        /* no results, figure out why and give a hopefully helpful hint. */
        bool have_disabled = false;
        for (GSimState::ship_index_t i = 0; i < sim_state.getNumShips(); ++i) {
            if ((sim_state.getShip(i).getFlags() & GSimObject::fl_Deactivated) != 0) {
                have_disabled = true;
                break;
            }
        }
        if (have_disabled) {
            messageBox(_("There are no fights in this arrangement. You have disabled "
                         "some ships, try re-enabling them using [.]."),
                       _("Simulator"));
        } else {
            messageBox(_("There are no fights in this arrangement. Did you set all "
                         "owners and aggression settings correctly?"),
                       _("Simulator"));
        }
    } else {
        /* okay, let's boogie */
        ui_root->centerWidget(result_window);
        int value = ui_root->runModal(result_window);
        if (value == cm_Enter) {
            int index = result_window.getResultListIndex();
            if (index >= int(sim_state.getNumShips()) && sim_state.hasPlanet())
                editPlanet();
            else
                ed_list.scrollTo(index, 1);
        }
    }
}

/** Write all ships to game (Ctrl-W). */
void
WSimScreen::writeToGame()
{
    /* only try when a game is available */
    if (!game_if.isGameAvailable())
        return;

    int count = 0, failed = 0;
    for (GSimState::ship_index_t i = 0; i < sim_state.getNumShips(); ++i) {
        const GSimShip& sh = sim_state.getShip(i);
        if (game_if.getShipRelation(sh) == GSimulatorGameInterface::ShipIsPlayed) {
            if (game_if.updateToGame(sh))
                ++count;
            else
                ++failed;
        }
    }

    if (count > 0)
        messageBox(format(_("%d ship%!1{s have%| has%} been written back to the game."), count),
                   _("Simulator"));
    else if (failed > 0)
        messageBox(format(_("%d ship%!1{s%} could not be written back to the game."), failed),
                   _("Simulator"));
    else
        messageBox(_("There are none of your ships in this simulation."),
                   _("Simulator"));
}

/** Update all ships from game (Ctrl-U). */
void
WSimScreen::updateFromGame()
{
    /* only try when a game is available */
    if (!game_if.isGameAvailable())
        return;

    int count = 0, failed = 0;
    for (GSimState::ship_index_t i = 0; i < sim_state.getNumShips(); ++i) {
        GSimShip& sh = sim_state.getShip(i);
        if (game_if.getShipRelation(sh) != GSimulatorGameInterface::ShipIsUnknown) {
            if (game_if.updateFromGame(sh, sim_opts))
                ++count;
            else
                ++failed;
        }
    }

    if (count > 0)
        messageBox(format(_("%d ship%!1{s have%| has%} been updated from the game."), count),
                   _("Simulator"));
    else if (failed > 0)
        messageBox(format(_("%d ship%!1{s%} could not be updated from the game."), failed),
                   _("Simulator"));
    else
        messageBox(_("There are none of your ships in this simulation."),
                   _("Simulator"));
}

bool
WSimScreen::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'p':
            editPlanet();
            sim_state.doUpdates();
            return true;
         case ss_Ctrl + 'w':
            writeToGame();
            return true;
         case ss_Ctrl + 'u':
            updateFromGame();
            sim_state.doUpdates();
            return true;
         case ss_Ctrl + 'o':
            editSimulatorOptions(sim_opts);
            sim_state.markStructureChange(); // config change might mean ship content changed (default hullfuncs)
            sim_state.doUpdates();
            return true;
         case ss_Ctrl + 'r':
            loadFile();
            return true;
         case ss_Ctrl + 's':
            saveFile();
            return true;
         case ss_Ctrl + 'c':
            doFleetCostComparison();
            return true;
         case ss_Ctrl + 'a':
            editAlliances(sim_opts, sim_default_relations, game_if);
            return true;
         case ss_Ctrl + SDLK_DELETE:
            if (UIMessageBox(_("Clear this simulation arrangement?"), _("Battle Simulator")).doYesNoDialog()) {
                GSimState().swap(sim_state);
                sim_state.doUpdates();
            }
            return true;
         case '\\':
         case '#':
            onContextMenu();
            return true;
         case SDLK_RETURN:
            runSimulation();
            return true;
         // Missing from PCC 1.x:
         //   F4
         // Missing from CCBSim 1.x:
         //   C-v (VCR view), C-l (load pconfig), M-i (version)
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
WSimScreen::onContextMenu()
{
    /* No context menu if we don't have anything on our list */
    if (sim_state.getNumShips() == 0)
        return;

    StringList sl;
    sl.add(0, _("Sort by Id"));
    sl.add(1, _("Sort by Owner"));
    sl.add(2, _("Sort by Hull"));
    sl.add(3, _("Sort by Battle Order"));
    sl.add(4, _("Sort by Name"));
    UIStandardListbox menu(0, 1, getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    if (menu.doStandardMenu(ed_list)) {
        switch (menu.getKey()) {
         case 0:
            sortBy(sortById);
            break;
         case 1:
            sortBy(sortByOwner);
            break;
         case 2:
            sortBy(sortByHull);
            break;
         case 3:
            sortBy(sortByBattleOrder);
            break;
         case 4:
            sortBy(sortByName);
            break;
        }
        sim_state.doUpdates();
    }
}

void
WSimScreen::onExtraMenu()
{
    StringList sl;
    sl.add(0, _("Options [Ctrl-O]"));
    sl.add(1, _("Alliances [Ctrl-A]"));
    if (sim_state.getNumShips() > 0) {
        sl.add(2, _("Fleet Cost Comparison [Ctrl-C]"));
    }
    UIStandardListbox menu(0, 1, getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    if (menu.doStandardMenu(btn_extras)) {
        switch (menu.getKey()) {
         case 0:
            editSimulatorOptions(sim_opts);
            break;
         case 1:
            editAlliances(sim_opts, sim_default_relations, game_if);
            break;
         case 2:
            doFleetCostComparison();
            break;
        }
    }
}

void
WSimScreen::sortBy(bool compare(const GSimShip& a, const GSimShip& b))
{
    /* Anything to do? */
    if (sim_state.getNumShips() == 0)
        return;

    /* Remember old selection */
    int oldId = sim_state.getShip(ed_list.getCurrent()).getId();

    /* Bubblesort. Who cares. */
    GSimState::ship_index_t n = 1;
    while (n < sim_state.getNumShips()) {
        if (compare(sim_state.getShip(n), sim_state.getShip(n-1))) {
            sim_state.swapShips(n, n-1);
            if (n > 1)
                --n;
            else
                ++n;
        } else {
            ++n;
        }
    }

    /* Scroll to new selection */
    ed_list.scrollTo(sim_state.getIndexOf(oldId), 1);
    sim_state.doUpdates();
}

void
initSimulatorState()
{
    // Discard old state
    doneSimulatorState();

    // Set up configuration
    WSimScreen::sim_opts.loadDefaults(!host.isPHost() ? GSimOptions::VcrHost :
                                      host.host_version < MKVERSION(3,0,0) ? GSimOptions::VcrPHost2 :
                                      host.host_version < MKVERSION(4,0,0) ? GSimOptions::VcrPHost3 : GSimOptions::VcrPHost4);
}

void
doneSimulatorState()
{
    WSimScreen::sim_opts = GSimOptions();
    GSimState().swap(WSimScreen::sim_state);
}

int
doSimulator(GSimulatorGameInterface& game_iface)
{
    if (WSimScreen::sim_state.hasPlanet()) {
        setPlanetName(game_iface, WSimScreen::sim_state.getPlanet());
    }

    WSimScreen me(game_iface);
    me.setExtent(ui_root->getExtent());
    return ui_root->runModal(me);
}

/***************************** Initialisation ****************************/

extern const ModuleInitializer mod_sim[] = {
    { OnEnteredGame, initSimulatorState },
    { OnLeftGame, doneSimulatorState },
    { 0, 0 },
};
