/**
  *  \file client/widgets/buildorderdisplay.cc
  *  \brief WBuildOrderDisplay widget
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/buildorderdisplay.h"
#include "cpluslib/format.h"
#include "game/spec.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "ui/util.h"

/** \class WBuildOrderDisplay
    \brief Display a Build Order

    This will display a build order on the ship build screen. */
WBuildOrderDisplay::WBuildOrderDisplay()
{
    o.hull = 0;
}

void
WBuildOrderDisplay::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin()).useFont(*fonts[FONT_NORMAL]);

    int x = getExtent().x;
    int y = getExtent().y;
    int w = getExtent().w;
    int dy = font_heights[FONT_NORMAL];

    drawBackground(ctx, getExtent());
    
    if (o.hull) {
        ctx.setColor(UIColor::tc_Contrast);
        outTextF(ctx, x, y, w, _("Your order:"));
        y += dy;
        ctx.setColor(UIColor::tc_Static);
        outTextFormatted(ctx, x, y, w, formatBuildOrder(o));
    }
    can.update(getExtent());
}

void
WBuildOrderDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(20, 5);
}

void
WBuildOrderDisplay::setBuildOrder(const TBuildOrder& o)
{
    this->o = o;
    drawWidget();
}


string_t
formatBuildOrder(const TBuildOrder& o)
{
    string_t result;
    if (o.hull) {
        const GHull& h = getHull(o.hull);
        result += "\xE2\x80\xA2 ";
        result += h.getName();
        result += "\n";

        /* These format strings turn into "Transwarp Drive" when there's one,
           "7 x Transwarp Drive" when there are many. */
        result += format(_("\xE2\x80\xA2 %!d%!1{%0$d \xC3\x97 %}%1$s"),
                         h.getNumEngines(),
                         getEngine(o.engine).getName());
        result += "\n";
        if (o.beam_count) {
            result += format(_("\xE2\x80\xA2 %!d%!1{%0$d \xC3\x97 %}%1$s"),
                             o.beam_count,
                             getBeam(o.beam_type).getName());
            result += "\n";
        }
        if (o.torp_count) {
            result += format(_("\xE2\x80\xA2 %!d%!1{%0$d \xC3\x97 %}%1$s"),
                             o.torp_count,
                             getTorp(o.torp_type).getName());
            result += "\n";
        }
        if (h.getNumBays()) {
            result += format(_("\xE2\x80\xA2 %d fighter bay%!1{s%}"), h.getNumBays());
            result += "\n";
        }
    }
    return result;
}
