/**
  *  \file client/widgets/specmaininfo.cc
  *  \brief Ship Specification Main Information
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/specmaininfo.h"
#include "ui/uiglobal.h"
#include "ui/table.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "util/misc.h"
#include "game/spec.h"
#include "game/host.h"
#include "game/hullfunc.h"
#include "game/global.h"
#include "game/config.h"
#include "resmgr/resmgr.h"
#include "gfx/palette.h"
#include "ui/util.h"
#include "game/pref.h"

static ConfigIntOption opt_hullfunc_images(getUserPreferences(), "Display.HullfuncImages", ValueBoolParser::instance, 1);

void
drawHullFunctionList(GfxContext& ctx,
                     const GfxRect& pos,
                     const GHullFunctionList& hflist,
                     const GShipQuery& query,
                     string_t prefix)
{
    if (opt_hullfunc_images()) {
        const int IMAGE_SIZE = 32;
        const int FRAME_EXTRA = 3;
        GfxRect area(pos);
        GfxRect line;
        ctx.useFont(*fonts[FONT_NORMAL]);

        for (size_t i = 0, n = hflist.size(); i < n; ++i) {
            const GHullFunction& hf = hflist[i];
            int damageLimit = 0;
            string_t image;
            if (const GBasicHullFunction* ptr = GHullFunctionData::instance.getBasicFunctionById(hf.getBasicFunctionId())) {
                damageLimit = ptr->getDamageLimit(query.getOwner(), config);
                image = ptr->getPictureName();
            }

            string_t prefix;
            uint8_t corner, text;
            bool playerOK = query.getPlayerDisplaySet().containsAnyOf(hf.getPlayers());
            bool levelOK = query.getLevelDisplaySet().containsAnyOf(hf.getLevels());
            if (playerOK && levelOK) {
                if (damageLimit >= 0 && query.getDamage() >= damageLimit) {
                    prefix = "abilities.r32.";
                    text = COLOR_RED;
                } else {
                    prefix = "abilities.32.";
                    text = COLOR_GRAY;
                }
                corner = 0;
            } else {
                prefix = "abilities.b32.";
                if (playerOK) {
                    corner = COLOR_DARKGREEN;
                    text = COLOR_YELLOW;
                } else {
                    corner = COLOR_RED;
                    text = COLOR_RED;
                }
            }

            Ptr<GfxPixmap> pix;
            if (!image.empty()) {
                pix = requestPixmap(ResId(prefix + image));
            }

            if (line.w < IMAGE_SIZE + FRAME_EXTRA) {
                drawBackground(ctx, line);
                line = area.splitY(IMAGE_SIZE + FRAME_EXTRA);
            }

            GfxRect pixArea = line.splitX(IMAGE_SIZE + FRAME_EXTRA);
            GfxCanvas& can = ctx.getCanvas(); 
            drawFrameDown(can, GfxRect(pixArea.x, pixArea.y, pixArea.w-1, pixArea.h-1));

            pixArea = GfxRect(pixArea.x + 1, pixArea.y + 1, pixArea.w - 3, pixArea.h - 3);
            drawSolidBar(can, pixArea, standard_colors[COLOR_BLACK]);
            if (pix.ptr() != 0) {
                pix->blit(ctx.getCanvas(), pixArea.x, pixArea.y, GfxRect(0, 0, IMAGE_SIZE, IMAGE_SIZE));
            } else {
                ctx.setTransparentBackground();
                ctx.setRawColor(standard_colors[text]);
                ctx.setTextAlign(1, 1);
                outTextF(ctx, pixArea, format("#%d", hf.getBasicFunctionId()));
            }

            if (corner != 0) {
                const int N = 5;
                for (int i = 0; i < N; ++i) {
                    can.drawHLine(GfxPoint(pixArea.x, pixArea.y+i), N-i, standard_colors[corner], SOLID_LINE, GFX_ALPHA_OPAQUE);
                }
            }
        }
        drawBackground(ctx, line);
        drawBackground(ctx, area);
    } else {
        int total = hflist.size();
        int nlines = pos.h / font_heights[FONT_NORMAL];
        int linesToDisplay = total > nlines ? nlines-1 : total;
        ctx.setSolidBackground();
        ctx.useFont(*fonts[FONT_NORMAL]);
        for (int i = 0; i < linesToDisplay; ++i) {
            ctx.setColor(UIColor::tc_Static);
            const GHullFunction& hf = hflist[i];

            // Figure out name
            string_t name;
            int damageLimit;
            if (const GBasicHullFunction* ptr = GHullFunctionData::instance.getBasicFunctionById(hf.getBasicFunctionId())) {
                name = ptr->getDescription();
                damageLimit = ptr->getDamageLimit(query.getOwner(), config);
            } else {
                name = format(_("Hull Function #%d"), hf.getBasicFunctionId());
                damageLimit = -1;
            }

            // Annotations for race/level restrictions
            string_t annot;
            if (damageLimit >= 0 && query.getDamage() >= damageLimit) {
                annot += _("damaged");
                ctx.setColor(UIColor::tc_Red);
            }

            string_t z = formatPlayerSet(hf.getPlayers());
            if (z.size()) {
                if (annot.size())
                    annot += "; ";
                annot += z;
            }
            z = formatExpLevelSet(hf.getLevels());
            if (z.size()) {
                if (annot.size())
                    annot += "; ";
                annot += z;
            }
            if (hf.getKind() == hf.AssignedToShip) {
                if (annot.size())
                    annot += "; ";
                annot += _("ship");
            }

            // Build total
            if (annot.size()) {
                name += " (";
                name += annot;
                name += ")";
            }

            // Display
            outTextF(ctx, pos.x, pos.y + i*font_heights[FONT_NORMAL], pos.w, prefix + name);
        }

        // Continuation
        if (linesToDisplay < total) {
            outTextF(ctx, pos.x, pos.y + linesToDisplay*font_heights[FONT_NORMAL], pos.w,
                     format(_("... (+%d more)"), total - linesToDisplay));
            ++linesToDisplay;
        }

        if (linesToDisplay < nlines) {
            drawBackground(ctx, GfxRect(pos.x, pos.y + linesToDisplay*font_heights[FONT_NORMAL],
                                        pos.w, (nlines - linesToDisplay)*font_heights[FONT_NORMAL]));
        }
    }
}

WSpecMainInfo::WSpecMainInfo()
{ }

WSpecMainInfo::~WSpecMainInfo()
{ }

void
WSpecMainInfo::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = GfxPoint(300, font_heights[FONT_NORMAL]*7 + 7);
    info.flags = li_FlexibleH;
}

void
WSpecMainInfo::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    if (query.getHullId() > 0) {
        int ntablelines = 3;
        UITableWriter tw(ctx,
                         GfxRect(getExtent().x, getExtent().y, getExtent().w, ntablelines*font_heights[FONT_NORMAL]),
                         FONT_NORMAL);

        /* Headings */
        tw.newLColumn();
        tw.writeFixedCell(UIColor::tc_Static, _("Weapons:"));
        tw.writeFixedCell(UIColor::tc_Static, _("Mine hit:"));
        tw.writeFixedCell(UIColor::tc_Static, _("Hull Id:"));
        if (host.isEugeneGame()) {
            tw.writeFixedCell(UIColor::tc_Static, _("Fuel burn:"));
        }

        /* Values */
        const GHull& h = getHull(query.getHullId());
        string_t w;
        if (h.getMaxBeams() > 0) {
            w += format(_("%d beam%!1{s%}"), h.getMaxBeams());
        }
        if (h.getMaxLaunchers() > 0) {
            if (w.size())
                w += ", ";
            w += format(_("%d torpedo launcher%!1{s%}"), h.getMaxLaunchers());
        }
        if (h.getNumBays() > 0) {
            if (w.size())
                w += ", ";
            w += format(_("%d fighter bay%!1{s%}"), h.getNumBays());
        }
        if (w.size() == 0)
            w = _("none");

        int mdamage;
        if (h.getMass() == 0)
            mdamage = 100;
        else if (host.isPHost())
            mdamage = 100 * config.MineHitDamageFor100KT() / h.getMass();
        else
            mdamage = radiv(10000, (h.getMass() + 1));

        tw.newLColumnUpTo(getExtent().getRightX());
        tw.writeVariableCell(UIColor::tc_Green, w);
        tw.writeVariableCell(UIColor::tc_Green, format(_("%d%% damage"), mdamage));
        tw.writeVariableCell(UIColor::tc_Green, numToString(query.getHullId()));
        if (host.isEugeneGame()) {
            tw.writeVariableCell(UIColor::tc_Green, format(_("%d kt/turn, %d kt/fight"),
                                                           getTurnFuelUsage(query.getHullId(), query.getOwner(), config.FuelUsagePerTurnFor100KT, 1),
                                                           getTurnFuelUsage(query.getHullId(), query.getOwner(), config.FuelUsagePerFightFor100KT, 1)));
            ++ntablelines;
        }

        tw.finish();

        /* Hull functions */
        GHullFunctionList hflist;
        if (haveDisplayedTurn()) {
            query.enumShipSpecials(&getDisplayedTurn().getCurrentUniverse(), hflist, false);
        } else {
            query.enumShipSpecials(0, hflist, false);
        }
        hflist.simplify();
        hflist.sortForNewShip(query.getPlayerDisplaySet());

        int nlines = getExtent().h / font_heights[FONT_NORMAL] - ntablelines;
        if (nlines > 0) {
            drawHullFunctionList(ctx,
                                 GfxRect(getExtent().x,
                                         getExtent().y + ntablelines*font_heights[FONT_NORMAL],
                                         getExtent().w,
                                         getExtent().h - ntablelines*font_heights[FONT_NORMAL]),
                                 hflist,
                                 query,
                                 string_t());
        }
    } else {
        drawBackground(ctx, getExtent());
    }
    can.update(getExtent());
}

void
WSpecMainInfo::setQuery(const GShipQuery& query)
{
    if (query != this->query) {
        this->query = query;
        drawWidget();
    }
}
