/**
  *  \file game/sim-io.cc
  *  \brief Simulator I/O
  *
  *  There are multiple versions of the .ccb file format. Here, we
  *  adopt the numbering used in the file format list, i.e. 0 =
  *  "CCsim", 1 = "CCbsim0", 2 = "CCbsim1" etc. Files differ in record
  *  sizes (more data for recent features added to the end), and in
  *  content (version <= 1 has only one starbase torp type).
  *
  *  We read all file formats, but only save a selection of the more
  *  recent ones. In this case, we generally use version 3, and use 4
  *  if its features (FLAK rating overrides) are required.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/sim.h"
#include "util/nls.h"
#include "game/struct.h"
#include "util/except.h"
#include "cpluslib/format.h"
#include "io/io.h"
#include "io/streams.h"

static const int MAX_VERSION  = 5;
static const int MAX_OBJ_SIZE = 67;

static const uint8 record_sizes[] = {
    51, 53, 53, 57, 65, 67
};

static const int MAGIC_V0_LEN = 6;
static const int MAGIC_V1_LEN = 8;

static const char magic_v0[] = { 'C','C','s','i','m',26 };
static const char magic_v1[] = { 'C','C','b','s','i','m','0',26 };

CTASSERT(MAX_OBJ_SIZE == TSimShipData::size);
CTASSERT(MAX_OBJ_SIZE == TSimPlanetData::size);

/** Load from file. Does not merge or ensure uniqueness of Id
    numbers.
    \param in Stream to read from */
void
GSimState::load(Stream& in)
{
    using std::memcmp;

    char magic[MAGIC_V1_LEN];
    int version;

    /* Check signature */
    in.readT(magic, MAGIC_V0_LEN);

    if (memcmp(magic, magic_v0, MAGIC_V0_LEN) == 0) {
        version = 0;
    } else if (memcmp(magic, magic_v1, MAGIC_V0_LEN) == 0) {
        in.readT(magic, 2);
        version = (magic[0] - '0') + 1;
        if (version <= 0 || version > MAX_VERSION)
            throw FileFormatException(in, _("Unsupported file format version"));
    } else {
        throw FileFormatException(in, _("File is missing required signature"));
    }

    /* Check object count */
    in.readT(magic, 2);

    uint16 count = getUint16(magic);
    bool has_planet = (count & 0x8000U) != 0;
    count &= 0x7FFF;

    /* Read objects */
    char obj_buf[MAX_OBJ_SIZE];
    for (unsigned int i = 0; i < count; ++i) {
        in.readT(obj_buf, record_sizes[version]);

        TSimShipData data;
        data.aggressiveness = -1;
        data.mass           = 100;
        data.flags          = 0;
        data.flak_r         = 0;
        data.flak_c         = 0;
        data.xa_id          = 0;
        getPartialStructure(obj_buf, &data, data.parser, 0, record_sizes[version]);

        getShip(addShip()) = GSimShip(data, version);
    }

    if (has_planet) {
        in.readT(obj_buf, record_sizes[version]);

        TSimPlanetData data;
        data.aggressiveness = -1;
        data._pad5          = 100;
        data.flags          = 0;
        data.flak_r         = 0;
        data.flak_c         = 0;
        data._pad6          = 0;
        getPartialStructure(obj_buf, &data, data.parser, 0, record_sizes[version]);

        addPlanet();
        getPlanet() = GSimPlanet(data, version);
    }
}

/** Check an object's version.
    \param o        [in] Object
    \param version  [in/out] Version, updated accordingly */
static void
checkVersion(GSimObject& o, int& version)
{
    int my_version = 3;
    if ((o.getFlags() >> 16) != 0) {
        my_version = 5;
    } else if (o.getFlags() & GSimObject::fl_RatingOverride) {
        my_version = 4;
    } else {
        GSimShip* sh = dynamic_cast<GSimShip*>(&o);
        if (sh != 0 && sh->getInterceptId() != 0)
            my_version = 4;
    }

    if (my_version > version)
        version = my_version;
}

/** Save to stream. Automatically picks the file version number.
    \param out stream to save to */
void
GSimState::save(Stream& out)
{
    using std::memcpy;

    /* Figure out what version to save in */
    int version = 3;
    for (ship_index_t i = 0; i < getNumShips(); ++i)
        checkVersion(getShip(i), version);
    if (hasPlanet())
        checkVersion(getPlanet(), version);

    /* Write header */
    char header[MAGIC_V1_LEN + 2];
    memcpy(header, magic_v1, MAGIC_V1_LEN);
    header[6] = '0' + version - 1;

    if (hasPlanet())
        storeUint16(header + MAGIC_V1_LEN, getNumShips() | 0x8000);
    else
        storeUint16(header + MAGIC_V1_LEN, getNumShips());

    out.writeT(header, sizeof(header));

    /* Write ships */
    char buffer[MAX_OBJ_SIZE];
    for (ship_index_t i = 0; i < getNumShips(); ++i) {
        TSimShipData data;
        getShip(i).storeData(data, version);
        storeStructure(buffer, data);
        out.writeT(buffer, record_sizes[version]);
    }

    /* Write planet */
    if (hasPlanet()) {
        TSimPlanetData data;
        getPlanet().storeData(data, version);
        storeStructure(buffer, data);
        out.writeT(buffer, record_sizes[version]);
    }
}
