/**
  *  \file game/trnapply.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/trnapply.h"
#include "io/io.h"
#include "arch/platform.h"
#include "util/recode.h"
#include "game/gamedirs.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/gen.h"
#include "game/msg.h"
#include "game/save.h"

static bool
applyTurnCommand(TShip& ship, const GTurnfile& f, const int i, string_t& fcAccum)
{
    if (f.getCommandType(i) == GTurnfile::ShipCommand && f.getCommandId(i) == ship.ship_id) {
        const int code = f.getCommandCode(i);
        getPartialStructure(f.getCommandData(i), &ship, ship.parser, f.getCommandCodeRecordIndex(code), f.getCommandLength(i));
        if (code == tcm_ShipChangeFc
            && i+1 < f.getNumCommands()
            && f.getCommandCode(i+1) == tcm_ShipChangeFc
            && f.getCommandId(i+1) == ship.ship_id)
        {
            // might be THost alliance command
            const char* data = f.getCommandData(i);
            const uint16_t hack = getUint16(data);
            if (hack == 0x6666 /* ff */ || hack == 0x4646 /* FF */ || hack == 0x6565 /* ee */) {
                fcAccum.append(data, 3);
            }
        }
        return true;
    } else {
        return false;
    }
}

static bool
applyTurnCommand(TPlanet& planet, const GTurnfile& f, const int i)
{
    if (f.getCommandType(i) == GTurnfile::PlanetCommand && f.getCommandId(i) == planet.planet_id) {
        const int code = f.getCommandCode(i);
        if (code == tcm_PlanetBuildBase) {
            planet.build_base = 1;
        } else {
            getPartialStructure(f.getCommandData(i), &planet, planet.parser, f.getCommandCodeRecordIndex(code), f.getCommandLength(i));
        }
        return true;
    } else {
        return false;
    }
}

static bool
applyTurnCommand(TStarbase& base, const GTurnfile& f, const int i)
{
    if (f.getCommandType(i) == GTurnfile::BaseCommand && f.getCommandId(i) == base.base_id) {
        const int code = f.getCommandCode(i);
        getPartialStructure(f.getCommandData(i), &base, base.parser, f.getCommandCodeRecordIndex(code), f.getCommandLength(i));
        return true;
    } else {
        return false;
    }
}

GTurnfileApply::GTurnfileApply(Ptr<GTurnfile> trn, bool use35)
    : trn(trn),
      use35(use35),
      cursor(0),
      cursor_obj(GTurnfile::UndefinedCommand),
      cursor_id(0)
{
    trn->sortCommands();
}

GTurnfileApply::~GTurnfileApply()
{ }

void
GTurnfileApply::updateShip(TShip& ship)
{
    const int id = ship.ship_id;
    moveCursor(GTurnfile::ShipCommand, id);
    while (cursor < trn->getNumCommands() && applyTurnCommand(ship, *trn, cursor, tim_allies)) {
        ++cursor;
    }
}

void
GTurnfileApply::updatePlanet(TPlanet& planet)
{
    moveCursor(GTurnfile::PlanetCommand, planet.planet_id);
    while (cursor < trn->getNumCommands() && applyTurnCommand(planet, *trn, cursor)) {
        ++cursor;
    }
}

void
GTurnfileApply::updateBase(TStarbase& base)
{
    moveCursor(GTurnfile::BaseCommand, base.base_id);
    while (cursor < trn->getNumCommands() && applyTurnCommand(base, *trn, cursor)) {
        ++cursor;
    }
}

void
GTurnfileApply::updateGen(GGen& gen)
{
    // Locate the last possible password-change command; this is the one that gets effective
    for (int i = trn->getNumCommands(); i > 0; --i) {
        if (trn->getCommandCode(i-1) == tcm_ChangePassword) {
            gen.setPasswordFromTurn(trn->getCommandData(i-1));
            break;
        }
    }
}

void
GTurnfileApply::finishDirectory()
{
    // Messages
    const int player = trn->getPlayer();
    GOutbox box(player);
    for (int i = 0, n = trn->getNumCommands(); i < n; ++i) {
        if (trn->getCommandCode(i) == tcm_SendMessage) {
            const char* data = trn->getCommandData(i);
            const int length = trn->getCommandId(i);
            const int recv = getUint16(data+2);
            if (recv > 0 && recv <= 12) {
                box.addMessageFromFile(getMessageFromArray(data+4, length, false), GPlayerSet(recv));
            }
        }
    }
    if (box.getCount() > 0) {
        if (use35) {
            saveOutbox35(box, *game_file_dir, player);
        } else {
            saveOutbox(box, *game_file_dir, player);
        }
    }

    // THost allies
    if (!tim_allies.empty()) {
        Ptr<Stream> cmdfile = game_file_dir->openFile(format("cmd%d.txt", trn->getPlayer()), Stream::C_CREATE | Stream::C_WRITE);
        string_t line =  _("# Additional commands");
        line += Platform::LINE_SEPARATOR;
        line += "$time ";
        line += trn->getTimestamp().getTimestampAsString();
        line += Platform::LINE_SEPARATOR;
        line += "$thost-allies ";
        line += tim_allies;
        line += Platform::LINE_SEPARATOR;
        line = convertUtf8ToGame(line);
        cmdfile->writeT(line.data(), line.size());
    }
}

/** Place cursor on command for object (obj,id).
    We don't want to rely on our caller calling updateShip/updatePlanet/updateBase
    in a specific order, neither do we want to loop through the turn file for each object.

    The common case is that all updateShip/updatePlanet/updateBase are called in ascending sequence,
    and GTurnfile::sortCommands() provides such a sequence for the commands.
    We therefore detect which sequence we are in, and provide amortized O(1) operation for those. */
void
GTurnfileApply::moveCursor(GTurnfile::CommandType obj, int id)
{
    // If the object type changes, or the id decreases, reposition the cursor from scratch.
    if (obj != cursor_obj || id <= cursor_obj) {
        cursor = 0;
        while (cursor < trn->getNumCommands() && trn->getCommandType(cursor) != obj) {
            ++cursor;
        }
    }

    // Locate first command that applies to our object
    while (cursor < trn->getNumCommands() && trn->getCommandType(cursor) == obj && trn->getCommandId(cursor) < id) {
        ++cursor;
    }

    // Remember where we were ordered
    cursor_obj = obj;
    cursor_id  = id;
}
