/**
  *  \file gfx/blit.cc
  *  \brief Stretch / Shear / Rotate Blit
  *
  *  This module contains operations for transformed blitting. We use
  *  a general approach to blit a rectangular pixmap to an arbitrary
  *  parallelogram.
  *
  *  \todo there's much that could be improved here: use a faster
  *  algorithm (some sort of Bresenham), anti-aliasing, etc.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cmath>
#include "gfx/blit.h"
#include "gfx/gfx-modetr.h"
#include "cpluslib/assert.h"

/** Back-end for stretching/shearing/rotating.
    See doStretchRotateBlit() for a description of the parameters.
    \param T framebuffer traits

    Implementation note: this uses the trick of computing the reverse
    transformation. That is, for each point in target_clip, we
    compute the corresponding original point. */
template<typename T>
inline void
doBlitInternal(SDL_Surface* src, SDL_Surface* target,
               const GfxRect& src_clip, const GfxRect& target_clip,
               int x, int y, int x1, int y1, int x2, int y2)
{
    const int det = (x2*y1 - y2*x1);
    if (det == 0)
        /* output axes are parallel or anti-parallel */
        return;

    for (int yp = 0, y_out = target_clip.y; yp < target_clip.h; ++yp, ++y_out) {
        uint8* p = T::add((uint8*) target->pixels + y_out * target->pitch, target_clip.x);
        for (int xp = 0, x_out = target_clip.x; xp < target_clip.w; ++xp, p = T::add(p, 1), ++x_out) {
            const int xpp = x_out - x;
            const int ypp = y_out - y;
            const int x0 = src_clip.w * (x2 * ypp - y2 * xpp) / det;
            const int y0 = src_clip.h * (y1 * xpp - x1 * ypp) / det;
            if (x0 < 0 || y0 < 0 || x0 >= src_clip.w || y0 >= src_clip.h)
                ;
            else
                T::poke(p, T::peek(T::add((uint8*) src->pixels + (y0 + src_clip.y) * src->pitch,
                                          x0 + src_clip.x)));
        }
    }
}

/** Simple, generic stretching/shearing/rotating routine. This
    function blits a transformed version of /src/ onto /target/. The
    input pixmap, a rectangle, will be mapped to a parallelogram which
    is given by two vectors.

    Note that both source and target canvases must be basic surfaces,
    i.e. not filters. To display a rotated pixmap on a UI canvas, you
    usually prepare it in a GfxPixmap object and blit that. This will
    also support proper clipping.

    \param src         source canvas
    \param target      target canvas
    \param src_clip    input rectangle (coordinates in terms of /src/)
    \param target_clip compute this area of /target/. Can be used to clip
                       output. Only this area is recomputed, so be sure to
                       make this big enough. On the other hand, don't make
                       it too big to not waste time.
    \param x,y         output origin. src's (0,0) will be transformed to
                       this position.
    \param x1,y1       vector to which X axis is mapped
    \param x2,y2       vector to which Y axis is mapped

    \pre surfaces have same pixel format. No conversion! */
void
doStretchRotateBlit(GfxCanvas& src, GfxCanvas& target,
                    GfxRect src_clip, GfxRect target_clip,
                    int x, int y, int x1, int y1, int x2, int y2)
{
    src_clip.clipRect(GfxRect(src));
    if (!src_clip.exists())
        /* input is empty */
        return;

    target_clip.clipRect(GfxRect(target));
    if (!target_clip.exists())
        /* input is empty */
        return;

    GfxSurface* src_sfc = dynamic_cast<GfxSurface*>(&src);
    GfxSurface* target_sfc = dynamic_cast<GfxSurface*>(&target);
    ASSERT(src_sfc != 0);
    ASSERT(target_sfc != 0);

    SDL_Surface* src_sdl = src_sfc->getSystemSurface();
    SDL_Surface* target_sdl = target_sfc->getSystemSurface();
    
    ASSERT(src_sdl->format->BitsPerPixel == target_sdl->format->BitsPerPixel);

#define sfc src_sdl
    GFX_MODE_SWITCH(doBlitInternal, (src_sdl, target_sdl, src_clip, target_clip, x, y, x1, y1, x2, y2));
#undef sfc
}

/** Compute bounding rectangle for stretch/rotate operation.
    \param x1,y1  X axis vector
    \param x2,y2  Y axis vector
    \return minimal rectangle enclosing the parallelogram defined by the parameters
    (minimum target_clip for a doStretchRotateBlit() anchored at (0,0). */
GfxRect
computeStretchRotateBBox(int x1, int y1, int x2, int y2)
{
    GfxRect rv;

    rv.x = std::min(x1, 0) + std::min(x2, 0);
    rv.y = std::min(y1, 0) + std::min(y2, 0);
    rv.w = std::abs(x1) + std::abs(x2);
    rv.h = std::abs(y1) + std::abs(y2);
    return rv;
}
