/**
  *  \file int/export/expdbf.cc
  *  \brief Export to DBF file
  *
  *  Unlike the text exporters, which use the type hints only as a guide to assign
  *  field widths, this one uses them to assign field types, and therefore require
  *  correctly-typed values to create a well-formatted file. So far, this restriction
  *  is implemented for bool, i.e. receiving a string or int instead of a bool will
  *  treat that with the getBoolValue() function, just like CCScript does when such
  *  a value is used in a boolean context.
  *
  *  This module is modelled after PCC 1.x export.pas::CDbfExporter.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <cstdlib>
#include "int/export/expdbf.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "int/value.h"
#include "io/io.h"
#include "int/interror.h"
#include "io/streams.h"

/** Constructor. */
IntDbfExporter::IntDbfExporter(Stream& s)
    : s(s),
      startPos(0),
      numRecords(0),
      recordSize(0)
{ }

/** Write dBASE file header. */
void
IntDbfExporter::writeFileHeader()
{
    char header[32];
    zeroFill(header);
    header[0] = 3;
    // header[1] = year;
    // header[2] = month;
    // header[3] = day;
    storeUint32(header+4, numRecords);
    storeUint16(header+8, startPos);
    storeUint16(header+10, recordSize);
    s.seek(0);
    s.writeT(header, sizeof(header));
}

void
IntDbfExporter::startTable(const IntExportFieldList& fields, const TIntTypeHint types[])
{
    // Initialize and write dummy header
    recordSize = 1;   /* for deletion marker */
    numRecords = 0;
    writeFileHeader();

    for (IntExportFieldList::index_t i = 0; i < fields.size(); ++i) {
        // Parse field definition
        int userWidth = std::abs(fields.getFieldWidth(i));
        int width;
        int decim;
        char typ;
        switch (types[i]) {
         case thBool:
            width = 1;
            decim = 0;
            typ = 'L';
            break;
            
         case thInt:
            width = userWidth ? userWidth : 10;
            decim = 0;
            typ = 'N';
            break;

         case thFloat:
            width = userWidth ? userWidth : 10;
            decim = 2;
            typ = 'N';
            break;

         case thString:
            width = userWidth ? userWidth : 30;
            decim = 0;
            typ = 'C';
            break;

         default:
            width = userWidth ? userWidth : 100;
            decim = 0;
            typ = 'C';
            break;
        }

        // Build field descriptor
        if (width > 255) {
            width = 255;
        }
        
        char fieldDesc[32];
        zeroFill(fieldDesc);
        std::strncpy(fieldDesc, fields.getFieldName(i).c_str(), 11);
        fieldDesc[11] = typ;
        fieldDesc[16] = width;
        fieldDesc[17] = decim;

        s.writeT(fieldDesc, sizeof(fieldDesc));

        widths.push_back(width);
        recordSize += width;

        // Validate. Refuse record sizes that cannot be represented in a 16-bit integer.
        // Refuse more than 1020 fields (which would mean the header size cannot be
        // represented in a 16-bit integer).
        if (recordSize >= 0x8000 || i >= 1020) {
            throw IntError("Too many fields for DBF export");
        }
    }

    // Write terminator
    s.writeT("\015", 1);
    startPos = s.getPos();

    record.assign(recordSize, char(0));
}

void
IntDbfExporter::startRecord()
{
    record[0] = ' ';
    recordPos = 1;
    fieldNum = 0;
}

void
IntDbfExporter::addField(IntValue* value, const string_t& /*name*/, TIntTypeHint type)
{
    if (type == thBool) {
        int bv = getBoolValue(value);
        if (bv == 0) {
            record[recordPos] = 'N';
        } else if (bv > 0) {
            record[recordPos] = 'Y';
        } else {
            record[recordPos] = '?';
        }
        ++recordPos;
    } else {
        string_t s;
        bool rightAlign;
        if (type == thFloat) {
            if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(value)) {
                s = format("%.2f", fv->getValue());
            } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(value)) {
                s = format("%d.00", iv->getValue());
            } else {
                /* treat as null */
            }
            rightAlign = true;
        } else if (type == thInt) {
            if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(value)) {
                s = format("%.0f", fv->getValue());
            } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(value)) {
                s = format("%d", iv->getValue());
            } else {
                /* treat as null */
            }
            rightAlign = true;
        } else {
            /* anything else */
            s = IntValue::toStringOf(value, false);
            rightAlign = false;
        }
        if (rightAlign && s.size() < widths[fieldNum]) {
            s.insert(s.begin(), widths[fieldNum] - s.size(), ' ');
        }
        storeBasicStringN(&record[recordPos], widths[fieldNum], s);
        recordPos += widths[fieldNum];
    }
    ++fieldNum;
}

void
IntDbfExporter::endRecord()
{
    s.writeT(&record[0], recordSize);
    ++numRecords;
}

void
IntDbfExporter::endTable()
{
    // Write one additional byte. My specs don't say this is needed,
    // but dbview.exe doesn't show the last record without it. This
    // may as well be a bug in dbview, but it doesn't hurt to have
    // this byte.
    s.writeT("", 1);
    writeFileHeader();
}
