/**
  *  \file u/t_int_keymap.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "u/t_int.h"
#include "int/keymap.h"
#include "ui/event.h"
#include "int/interror.h"

void
IntKeymapTestSuite::testParse()
{
    // Test parseKey
    TS_ASSERT_EQUALS(parseKey("a"),       uint32_t('a'));
    TS_ASSERT_EQUALS(parseKey("s-a"),     uint32_t('A'));
    TS_ASSERT_EQUALS(parseKey("A"),       uint32_t('a'));
    TS_ASSERT_EQUALS(parseKey("s-A"),     uint32_t('A'));
    TS_ASSERT_EQUALS(parseKey("shift-A"), uint32_t('A'));
    TS_ASSERT_EQUALS(parseKey("c-A"),     uint32_t(ss_Ctrl + 'a'));
    TS_ASSERT_EQUALS(parseKey("a-A"),     uint32_t(ss_Alt + 'a'));
    TS_ASSERT_EQUALS(parseKey("c-a-A"),   uint32_t(ss_Ctrl + ss_Alt + 'a'));
    TS_ASSERT_EQUALS(parseKey("a-c-A"),   uint32_t(ss_Ctrl + ss_Alt + 'a'));
    TS_ASSERT_EQUALS(parseKey("alt-ctrl-A"), uint32_t(ss_Ctrl + ss_Alt + 'a'));
    TS_ASSERT_EQUALS(parseKey("alt-c-A"), uint32_t(ss_Ctrl + ss_Alt + 'a'));
    TS_ASSERT_EQUALS(parseKey("A-A-A"),   uint32_t(ss_Alt + 'a'));

    TS_ASSERT_EQUALS(parseKey("1"),       uint32_t('1'));
    TS_ASSERT_EQUALS(parseKey("s-1"),     uint32_t(ss_Shift + '1'));

    TS_ASSERT_EQUALS(parseKey("f1"),      uint32_t(SDLK_F1));
    TS_ASSERT_EQUALS(parseKey("c-f1"),    uint32_t(ss_Ctrl + SDLK_F1));
    TS_ASSERT_EQUALS(parseKey("s-f1"),    uint32_t(ss_Shift + SDLK_F1));
    TS_ASSERT_EQUALS(parseKey("m-f1"),    uint32_t(ss_Meta + SDLK_F1));
    TS_ASSERT_EQUALS(parseKey("a-f1"),    uint32_t(ss_Alt + SDLK_F1));
    TS_ASSERT_EQUALS(parseKey("c-a-f1"),  uint32_t(ss_Ctrl + ss_Alt + SDLK_F1));
    TS_ASSERT_EQUALS(parseKey("s-a-f1"),  uint32_t(ss_Shift + ss_Alt + SDLK_F1));

    TS_ASSERT_EQUALS(parseKey("wheelup"), uint32_t(vk_WheelUp));
    TS_ASSERT_EQUALS(parseKey("WheelUp"), uint32_t(vk_WheelUp));
    TS_ASSERT_EQUALS(parseKey("BS"),      uint32_t(SDLK_BACKSPACE));
    TS_ASSERT_EQUALS(parseKey("ctrl-h"),  uint32_t(ss_Ctrl + 'h'));
    TS_ASSERT_EQUALS(parseKey("ctrl-m"),  uint32_t(ss_Ctrl + 'm'));
    TS_ASSERT_EQUALS(parseKey("ctrl-j"),  uint32_t(ss_Ctrl + 'j'));
    TS_ASSERT_EQUALS(parseKey("enter"),   uint32_t(SDLK_RETURN));
    TS_ASSERT_EQUALS(parseKey("ret"),     uint32_t(SDLK_RETURN));
    TS_ASSERT_EQUALS(parseKey("ctrl-enter"), uint32_t(ss_Ctrl + SDLK_RETURN));
}

void
IntKeymapTestSuite::testFormat()
{
    // Test formatKey
    TS_ASSERT_EQUALS(formatKey('a'),                         "A");
    TS_ASSERT_EQUALS(formatKey('A'),                         "SHIFT-A");
    TS_ASSERT_EQUALS(formatKey(ss_Ctrl + 'a'),               "CTRL-A");
    TS_ASSERT_EQUALS(formatKey(ss_Alt + 'a'),                "ALT-A");
    TS_ASSERT_EQUALS(formatKey(ss_Ctrl + ss_Alt + 'a'),      "CTRL-ALT-A");
    TS_ASSERT_EQUALS(formatKey('1'),                         "1");
    TS_ASSERT_EQUALS(formatKey(ss_Shift + '1'),              "SHIFT-1");
    TS_ASSERT_EQUALS(formatKey(SDLK_F1),                     "F1");
    TS_ASSERT_EQUALS(formatKey(ss_Ctrl + SDLK_F1),           "CTRL-F1");
    TS_ASSERT_EQUALS(formatKey(ss_Shift + SDLK_F1),          "SHIFT-F1");
    TS_ASSERT_EQUALS(formatKey(ss_Meta + SDLK_F1),           "META-F1");
    TS_ASSERT_EQUALS(formatKey(ss_Alt + SDLK_F1),            "ALT-F1");
    TS_ASSERT_EQUALS(formatKey(ss_Ctrl + ss_Alt + SDLK_F1),  "CTRL-ALT-F1");
    TS_ASSERT_EQUALS(formatKey(ss_Shift + ss_Alt + SDLK_F1), "SHIFT-ALT-F1");
    TS_ASSERT_EQUALS(formatKey(vk_WheelUp),                  "WHEELUP");
    TS_ASSERT_EQUALS(formatKey(SDLK_BACKSPACE),              "BS");
    TS_ASSERT_EQUALS(formatKey(ss_Ctrl + 'h'),               "CTRL-H");
    TS_ASSERT_EQUALS(formatKey(ss_Ctrl + 'm'),               "CTRL-M");
    TS_ASSERT_EQUALS(formatKey(ss_Ctrl + 'j'),               "CTRL-J");
    TS_ASSERT_EQUALS(formatKey(SDLK_RETURN),                 "RET");
    TS_ASSERT_EQUALS(formatKey(ss_Ctrl + SDLK_RETURN),       "CTRL-RET");
}

void
IntKeymapTestSuite::testKeymap()
{
    IntKeymapRef a, b;
    initKeymaps();

    // Create TESTKEYMAP
    TS_ASSERT_EQUALS(IntKeymap::getKeymapByName("TESTKEYMAP"), IntKeymapRef(0));
    a = IntKeymap::createKeymap("TESTKEYMAP");
    TS_ASSERT_EQUALS(IntKeymap::getKeymapByName("TESTKEYMAP"), a);
    TS_ASSERT_THROWS(IntKeymap::createKeymap("TESTKEYMAP"), IntError);

    // Create TESTCHILD
    b = IntKeymap::createKeymap("TESTCHILD");
    TS_ASSERT_EQUALS(IntKeymap::getKeymapByName("TESTCHILD"), b);

    // Check parents
    TS_ASSERT(!a->hasParent(b));
    TS_ASSERT(!b->hasParent(a));
    TS_ASSERT(a->hasParent(a));
    TS_ASSERT(b->hasParent(b));
    TS_ASSERT_THROWS_NOTHING(b->addParent(a));
    TS_ASSERT(!a->hasParent(b));
    TS_ASSERT(b->hasParent(a));
    TS_ASSERT_THROWS(b->addParent(a), IntError);
    TS_ASSERT_THROWS(a->addParent(b), IntError);
    TS_ASSERT_THROWS(a->addParent(a), IntError);
    TS_ASSERT_THROWS(b->addParent(b), IntError);

    // Check keys
    a->addKey(1, 2, 3);
    a->addKey(4, 5, 6);
    b->addKey(1, 4, 5);
    b->addKey(7, 8, 9);
    TS_ASSERT_EQUALS(a->lookupCommand(1), 2U);
    TS_ASSERT_EQUALS(a->lookupCommand(4), 5U);
    TS_ASSERT_EQUALS(a->lookupCommand(7), 0U);
    TS_ASSERT_EQUALS(a->lookupCommand(99), 0U);
    TS_ASSERT_EQUALS(b->lookupCommand(1), 4U);
    TS_ASSERT_EQUALS(b->lookupCommand(4), 5U);
    TS_ASSERT_EQUALS(b->lookupCommand(7), 8U);
    TS_ASSERT_EQUALS(b->lookupCommand(99), 0U);

    doneKeymaps();
}
